#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2021/3/27 14:18:09

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import os
import os.path as opath
from pathlib import Path
import shutil
from contextlib import contextmanager
#===============================================================================
# 
#===============================================================================

@contextmanager
def scope_work_dir(dir):
    try:
        old_dir = os.getcwd()
        yield os.chdir(dir)
    finally:
        os.chdir(old_dir)

def path_native(fpath):
    return opath.normpath(fpath).replace('\\', '/').replace('/', os.sep)

def path_unix(fpath):
    return opath.normpath(fpath).replace('\\', '/')

def path_win(fpath):
    return opath.normpath(fpath).replace('/', '\\')

def path_rel(fpath, parent):
    return path_unix(os.path.relpath(fpath, parent))

def files_env(key, filename):
    for pth in os.getenv(key).split(';'):
        full_path = opath.join(pth, filename)
        if opath.exists(full_path):
            yield path_unix(full_path)

def file_env(key, filename):
    for pth in os.getenv('path').split(';'):
        full_path = opath.join(pth, filename)
        if opath.exists(full_path):
            return path_unix(full_path)

def path_exe_file(filename):
    if opath.exists(filename):
        return path_unix(filename)
    return file_env('path', filename)

def path_exe_files(filename):
    if opath.exists(filename):
        yield path_unix(filename)
    for full_path in files_env('path', filename):
        yield path_unix(full_path)

def path_parent(fpath):
    '''fpath: 输入路径名字
       ret: 返回父路径
    '''
    return path_unix(opath.split(fpath)[0])

def path_make_parent(fpath):
    parent = path_parent(fpath)
    os.makedirs(parent, exist_ok=True)
    return parent

def path_make(fpath):
    os.makedirs(fpath, exist_ok=True)

def path_join(*args, make_parent=False):
    fpath = os.path.join(*args)
    if make_parent:
        path_make_parent(fpath)
    return path_unix(fpath)

def path_abs(fpath):
    return path_unix(os.path.abspath(fpath))

def path_exist(fpath):
    return fpath and os.path.exists(fpath)

def file_size(fpath):
    return os.path.getsize(fpath)

def file_delete(fpath):
    os.remove(fpath)

def dir_delete(dirpath):
    shutil.rmtree(dirpath)

def path_is_file(fpath):
    return fpath and os.path.isfile(fpath)

def path_is_dir(fpath):
    return fpath and os.path.isdir(fpath)

def path_filename(fpath, level=0):
    '''fpath: 输入路径名字
       ret: 返回文件名字
    '''
    splits = Path(fpath).splitall()
    return opath.split(fpath)[1]

def path_basename(fpath):
    '''file_path: 输入路径名字
       ret: 返回去掉后缀的文件名字
    '''
    return os.path.basename(fpath)

def path_sterm(fpath):
    return Path(fpath).stem

def path_ext(fpath, need_dot = True):
    '''file_path: 输入路径名字
       ret: 返回后缀
    '''
    fname = opath.split(fpath)[1]
    ext = opath.splitext(fname)[1]
    if ext and not need_dot:
        return ext[1:]
    return ext

def path_replace_ext(file_path, ext):
    '''file_path: 输入路径名字
       ext: 需要替换的后缀名字
       ret: 返回替换后的路径
    '''
    if isinstance(ext, str) and not ext.startswith('.'):
        ext = '.' + ext
    new_file = opath.splitext(file_path)[0] + ext
    return path_unix(new_file)

def path_replace_stem(file_path, basename):
    '''
        file_path：输入路径名字
        basename: 替换的文件名字,保留后缀
        ret: 返回替换后的路径
    '''
    _dir, _fname = opath.split(file_path)
    _basename, _ext = opath.splitext(_fname)
    return path_unix(opath.join(_dir, basename + _ext))

def path_replace_basename(file_path, filename):
    '''
        file_path：输入路径名字
        filename: 替换的文件名字
        ret: 返回替换后的路径
    '''
    _dir, _fname = opath.split(file_path)
    return path_unix(opath.join(_dir, filename))

def path_replace_parent(file_path, parent):
    '''
        file_path：输入路径名字
        parent: 替换的父路径
        ret: 返回替换后的路径
    '''
    _dir, _fname = opath.split(file_path)
    return path_unix(opath.join(parent, _fname))

def _ensure_list(suffix_list):
    if suffix_list is None:
        return []
    if isinstance(suffix_list, list):
        return suffix_list
    if isinstance(suffix_list, str):
        return suffix_list.split(';')
    return suffix_list

def _file_name_endswith(fname, suffix_list):
    return suffix_list and any( (fname.lower().endswith(s.lower()) for s in suffix_list) )

def files_under(dir_name, suffix_list = None, need_join_dir=True):
    '''
    dir_name: 路径
    suffix_list: 后缀列表, 比如: [.txt,.py];用于筛选文件
    ret: 返回生成器
    '''
    suffix_list = _ensure_list(suffix_list)
    for fdir_name, _sub_fdirs, files in os.walk(dir_name):
        for fname in files:
            if not suffix_list or _file_name_endswith(fname, suffix_list):
                yield path_unix(os.path.join(fdir_name, fname) if need_join_dir else fname)

def files_current(dir_name, suffix_list = None, need_join_dir=True):
    '''
    dir_name: 路径
    suffix_list: 后缀列表, 比如: [.txt,.py];用于筛选文件
    ret: 返回生成器
    '''
    suffix_list = _ensure_list(suffix_list)
    for fname in os.listdir(dir_name):
        fpath = os.path.join(dir_name, fname)
        if os.path.isfile(fpath):
            if not suffix_list or _file_name_endswith(fname, suffix_list):
                yield path_unix(fpath if need_join_dir else fname)

def dirs_under(dir_name, suffix_list = None, need_join_dir=True):
    '''
    dir_name: 路径
    suffix_list: 后缀列表, 比如: [.txt,.py]
    ret: 返回生成器
    '''
    suffix_list = _ensure_list(suffix_list)
    for fdir_name, sub_fdirs, files in os.walk(dir_name):
        for fname in sub_fdirs:
            if not suffix_list or _file_name_endswith(fname, suffix_list):
                yield path_unix(os.path.join(fdir_name, fname) if need_join_dir else fname)

def dirs_current(dir_name, suffix_list = None, need_join_dir=True):
    '''
    dir_name: 路径
    suffix_list: 后缀列表, 比如: [test, test2];用于筛选路径
    ret: 返回生成器
    '''
    suffix_list = _ensure_list(suffix_list)
    for fname in os.listdir(dir_name):
        fpath = os.path.join(dir_name, fname)
        if os.path.isdir(fpath):
            if not suffix_list or _file_name_endswith(fname, suffix_list):
                yield path_unix(fpath if need_join_dir else fname)




#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/3/7 15:50:28

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import os
import json
import numpy as np
from torch.utils.data import Dataset
from laok.cv3d.trans.trans import pc_normalize
#===============================================================================
# 
#===============================================================================

class ShapeNetCorePartNormal(Dataset):
    def __init__(self, root, npoints=2500, split='train', class_choice=None, normal_channel=True):
        assert (split in ["trainval", "train", "val", "test"])

        self.npoints = npoints
        self.root = root
        self.catfile = os.path.join(self.root, 'synsetoffset2category.txt')
        self.cat = {}
        self.normal_channel = normal_channel

        with open(self.catfile, 'r') as f:
            for line in f:
                ls = line.strip().split()
                self.cat[ls[0]] = ls[1]
        # self.cat = {k: v for k, v in self.cat.items()}
        self.classes_original = dict(zip(self.cat, range(len(self.cat))))

        if class_choice is not None:
            self.cat = {k:v for k,v in self.cat.items() if k in class_choice}
        # print(self.cat)

        print(f'catogery:{self.cat}')
        print(f'classes_original:{self.classes_original}')

        self.meta = {}
        with open(os.path.join(self.root, 'train_test_split', 'shuffled_train_file_list.json'), 'r') as f:
            train_ids = set([str(d.split('/')[2]) for d in json.load(f)])
        with open(os.path.join(self.root, 'train_test_split', 'shuffled_val_file_list.json'), 'r') as f:
            val_ids = set([str(d.split('/')[2]) for d in json.load(f)])
        with open(os.path.join(self.root, 'train_test_split', 'shuffled_test_file_list.json'), 'r') as f:
            test_ids = set([str(d.split('/')[2]) for d in json.load(f)])
        for item in self.cat:
            # print('category', item)
            self.meta[item] = []
            dir_point = os.path.join(self.root, self.cat[item])
            fns = sorted(os.listdir(dir_point))
            # print(fns[0][0:-4])
            if split == 'trainval':
                fns = [fn for fn in fns if ((fn[0:-4] in train_ids) or (fn[0:-4] in val_ids))]
            elif split == 'train':
                fns = [fn for fn in fns if fn[0:-4] in train_ids]
            elif split == 'val':
                fns = [fn for fn in fns if fn[0:-4] in val_ids]
            elif split == 'test':
                fns = [fn for fn in fns if fn[0:-4] in test_ids]
            else:
                print('Unknown split: %s. Exiting..' % (split))
                exit(-1)

            # print(os.path.basename(fns))
            for fn in fns:
                token = (os.path.splitext(os.path.basename(fn))[0])
                self.meta[item].append(os.path.join(dir_point, token + '.txt'))

        self.datapath = []
        for item in self.cat:
            for fn in self.meta[item]:
                self.datapath.append((item, fn))

        print('The size of %s data is %d'%(split,len(self.datapath)))

        self.classes = {}
        for i in self.cat.keys():
            self.classes[i] = self.classes_original[i]

        # Mapping from category ('Chair') to a list of int [10,11,12,13] as segmentation labels
        self.seg_classes = {'Earphone': [16, 17, 18], 'Motorbike': [30, 31, 32, 33, 34, 35], 'Rocket': [41, 42, 43],
                            'Car': [8, 9, 10, 11], 'Laptop': [28, 29], 'Cap': [6, 7], 'Skateboard': [44, 45, 46],
                            'Mug': [36, 37], 'Guitar': [19, 20, 21], 'Bag': [4, 5], 'Lamp': [24, 25, 26, 27],
                            'Table': [47, 48, 49], 'Airplane': [0, 1, 2, 3], 'Pistol': [38, 39, 40],
                            'Chair': [12, 13, 14, 15], 'Knife': [22, 23]}

        # for cat in sorted(self.seg_classes.keys()):
        #     print(cat, self.seg_classes[cat])

        self.cache = {}  # from index to (point_set, cls, seg) tuple
        self.cache_size = 20000


    def __getitem__(self, index):
        if index in self.cache:
            point_set, cls, seg = self.cache[index]
        else:
            fn = self.datapath[index]
            cat = self.datapath[index][0]
            cls = self.classes[cat]
            cls = np.array([cls]).astype(np.int32)
            data = np.loadtxt(fn[1]).astype(np.float32)
            if not self.normal_channel:
                point_set = data[:, 0:3]
            else:
                point_set = data[:, 0:6]
            seg = data[:, -1].astype(np.int32)
            if len(self.cache) < self.cache_size:
                self.cache[index] = (point_set, cls, seg)
        point_set[:, 0:3] = pc_normalize(point_set[:, 0:3])

        choice = np.random.choice(len(seg), self.npoints, replace=True)
        # resample
        point_set = point_set[choice, :]
        seg = seg[choice]

        return point_set, cls, seg

    def __len__(self):
        return len(self.datapath)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/4/19 13:51:55

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import threading
import concurrent.futures as futures
from .fpath import (files_under as _files_under,
                    files_current as _files_current,
                    dirs_current as _dirs_current,
                    dirs_under as _dirs_under)
#===============================================================================
'''     
'''
#===============================================================================

####################    线城接口
def thread_id():
    return threading.get_ident()

def thread_count():
    return threading.active_count()

####################    并行接口
def parallel_sequence(data_list, handler, max_workers=None, use_thread = True):
    if use_thread:
        Executor = futures.ThreadPoolExecutor
    else:
        Executor = futures.ProcessPoolExecutor

    with Executor(max_workers=max_workers) as ex:
        wait_for = [ex.submit(handler, data) for data in data_list]
        return [f.result() for f in futures.as_completed(wait_for)]

def files_under(dir_name, handler, suffix_list = None, need_join_dir=True, max_workers=None, use_thread = True):
    '''
    '''
    data_iter = _files_under(dir_name, suffix_list, need_join_dir)
    return parallel_sequence(data_iter, handler, max_workers=max_workers, use_thread=use_thread)

def files_current(dir_name, handler, suffix_list = None, need_join_dir=True, max_workers=None, use_thread = True):
    '''
    '''
    data_iter = _files_current(dir_name, suffix_list, need_join_dir)
    return parallel_sequence(data_iter, handler, max_workers=max_workers, use_thread=use_thread)

def dirs_under(dir_name, handler, suffix_list = None, need_join_dir=True, max_workers=None, use_thread = True):
    '''
    '''
    data_iter = _dirs_under(dir_name, suffix_list, need_join_dir)
    return parallel_sequence(data_iter, handler, max_workers=max_workers, use_thread=use_thread)

def dirs_current(dir_name, handler, suffix_list = None, need_join_dir=True, max_workers=None, use_thread = True):
    '''
    '''
    data_iter = _dirs_current(dir_name, suffix_list, need_join_dir)
    return parallel_sequence(data_iter, handler, max_workers=max_workers, use_thread=use_thread)





#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/3/7 15:43:22

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import os
# import json
# import torch
# import random
import numpy as np
from torch.utils.data import Dataset
from laok.cv3d.trans import farthest_point_sample, pc_normalize
# import laok.cv2d as kcv2
# import cv2
#===============================================================================
# 
#===============================================================================
__all__ = ['ModelNetNormalResampled']

class ModelNetNormalResampled(Dataset):
    def __init__(self, root, npoint=1024, split='train', num_category=40, uniform=False, normal_channel=True, cache_size=1):
        assert (split == 'train' or split == 'test')
        self.root = root
        self.npoints = npoint
        self.uniform = uniform
        self.normal_channel = normal_channel

        # 加载类别列表
        catfile = os.path.join(self.root, f'modelnet{num_category}_shape_names.txt')
        with open(catfile) as f:
            self.classes = {line.rstrip():i for i,line in enumerate(f)}
        print(f'classes : {self.classes}')

        # 加载数据文件列表
        shape_name_file = f'modelnet{num_category}_{split}.txt'
        print(f'{split} file : {shape_name_file}')
        with open(os.path.join(self.root, shape_name_file) )as f:
            shape_ids = [line.rstrip() for line in f]
        shape_names = ['_'.join(x.split('_')[0:-1]) for x in shape_ids] #去除尾数
        # list of (shape_name, shape_txt_file_path) tuple
        self.datapath = [(shape_names[i], os.path.join(self.root, shape_names[i], shape_ids[i]) + '.txt') for i
                         in range(len(shape_ids))]
        print(f'The size of {split} data is {len(self.datapath)}')

        self.cache_size = cache_size  # how many data points to cache in memory
        self.cache = {}  # from index to (point_set, cls) tuple

    def __len__(self):
        return len(self.datapath)

    def __getitem__(self, index):
        if index in self.cache:
            point_set, cls = self.cache[index]
        else:
            name, dfile = self.datapath[index]
            cls = self.classes[name]
            cls = np.array([cls]).astype(np.int32)
            point_set = np.loadtxt(dfile, delimiter=',').astype(np.float32)

            if self.uniform:
                point_set = farthest_point_sample(point_set, self.npoints)
            else:
                point_set = point_set[0:self.npoints,:]

            point_set[:, 0:3] = pc_normalize(point_set[:, 0:3])

            if not self.normal_channel:
                point_set = point_set[:, 0:3]

            if len(self.cache) < self.cache_size:
                self.cache[index] = (point_set, cls)

        return point_set, cls




#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/3/7 15:43:22

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import os
import json
import torch
import random
import numpy as np
from torch.utils.data import Dataset
from laok.cv3d.trans import farthest_point_sample, pc_normalize
# import laok.cv2d as kcv2
# import cv2
#===============================================================================
# 
#===============================================================================
__all__ = ['ModelNetImage', 'ModelNetImageFloatBin', 'ModelNetNormalResampled']

class ModelNetNormalResampled(Dataset):
    def __init__(self, root, npoint=1024, split='train', num_category=40, uniform=False, normal_channel=True, cache_size=15000):
        assert (split == 'train' or split == 'test')
        self.root = root
        self.npoints = npoint
        self.uniform = uniform
        self.normal_channel = normal_channel

        # 加载类别列表
        catfile = os.path.join(self.root, f'modelnet{num_category}_shape_names.txt')
        with open(catfile) as f:
            self.classes = {line.rstrip():i for i,line in enumerate(f)}
        print(f'classes : {self.classes}')

        # 加载数据文件列表
        shape_name_file = f'modelnet{num_category}_{split}.txt'
        print(f'{split} file : {shape_name_file}')
        with open(os.path.join(self.root, shape_name_file) )as f:
            shape_ids = [line.rstrip() for line in f]
        shape_names = ['_'.join(x.split('_')[0:-1]) for x in shape_ids] #去除尾数
        # list of (shape_name, shape_txt_file_path) tuple
        self.datapath = [(shape_names[i], os.path.join(self.root, shape_names[i], shape_ids[i]) + '.txt') for i
                         in range(len(shape_ids))]
        print(f'The size of {split} data is {len(self.datapath)}')

        self.cache_size = cache_size  # how many data points to cache in memory
        self.cache = {}  # from index to (point_set, cls) tuple

    def __len__(self):
        return len(self.datapath)

    def __getitem__(self, index):
        if index in self.cache:
            point_set, cls = self.cache[index]
        else:
            name, dfile = self.datapath[index]
            cls = self.classes[name]
            cls = np.array([cls]).astype(np.int32)
            point_set = np.loadtxt(dfile, delimiter=',').astype(np.float32)

            if self.uniform:
                point_set = farthest_point_sample(point_set, self.npoints)
            else:
                point_set = point_set[0:self.npoints,:]

            point_set[:, 0:3] = pc_normalize(point_set[:, 0:3])

            if not self.normal_channel:
                point_set = point_set[:, 0:3]

            if len(self.cache) < self.cache_size:
                self.cache[index] = (point_set, cls)

        return point_set, cls

class ModelNetImage(Dataset):
    def __init__(self, root, img_size=(200, 200), split='train', num_category=40, transform = None):
        assert (split == 'train' or split == 'test')
        self.root = root

        # 加载类别列表
        catfile = os.path.join(self.root, f'modelnet{num_category}_shape_names.txt')
        with open(catfile) as f:
            self.classes = {line.rstrip(): i for i, line in enumerate(f)}
        print(f'classes : {self.classes}')

        # 加载数据文件列表
        shape_name_file = f'modelnet{num_category}_{split}.txt'
        print(f'{split} file : {shape_name_file}')
        with open(os.path.join(self.root, shape_name_file))as f:
            shape_ids = [line.rstrip() for line in f]
        shape_names = ['_'.join(x.split('_')[0:-1]) for x in shape_ids]  # 去除尾数
        # list of (shape_name, shape_txt_file_path) tuple
        self.datapath = [(shape_names[i], os.path.join(self.root, shape_names[i], shape_ids[i])) for i
                         in range(len(shape_ids))]
        print(f'The size of {split} data is {len(self.datapath)}')
        self.img_size = img_size
        self.transform = transform

    def __len__(self):
        return len(self.datapath)

    def __getitem__(self, index):
        name, dfile = self.datapath[index]
        cls = self.classes[name]
        cls = np.array([cls]).astype(np.int32)
        xImgFile = dfile + '-X.jpg'
        yImgFile = dfile + '-Y.jpg'
        zImgFile = dfile + '-Z.jpg'

        xImg = kcv2.read_cv_gray(xImgFile)
        yImg = kcv2.read_cv_gray(yImgFile)
        zImg = kcv2.read_cv_gray(zImgFile)

        width, height = self.img_size
        xImg = kcv2.resize_abs(xImg, width, height)
        yImg = kcv2.resize_abs(yImg, width, height)
        zImg = kcv2.resize_abs(zImg, width, height)
        img = cv2.merge([xImg, yImg, zImg])

        img = kcv2.cv2pil(img)
        if self.transform:
            img = self.transform(img)
        # img2 = kcv2.hconcat([xImg, yImg, zImg])
        # kcv2.show(img2, 'img2')
        # kcv2.show(xImg, 'xImg')
        # kcv2.show(yImg, 'yImg')
        # kcv2.show(zImg, 'zImg')
        # kcv2.show_wait(img)
        return img, cls

class ModelNetImageFloatBin(Dataset):
    def __init__(self, root, img_size=(100, 100), split='train', num_category=40, transform = None, cache_size=500):
        assert (split == 'train' or split == 'test')
        self.root = root

        # 加载类别列表
        catfile = os.path.join(self.root, f'modelnet{num_category}_shape_names.txt')
        with open(catfile) as f:
            self.classes = {line.rstrip(): i for i, line in enumerate(f)}
        print(f'classes : {self.classes}')

        # 加载数据文件列表
        shape_name_file = f'modelnet{num_category}_{split}.txt'
        print(f'{split} file : {shape_name_file}')
        with open(os.path.join(self.root, shape_name_file))as f:
            shape_ids = [line.rstrip() for line in f]
        shape_names = ['_'.join(x.split('_')[0:-1]) for x in shape_ids]  # 去除尾数
        # list of (shape_name, shape_txt_file_path) tuple
        self.datapath = [(shape_names[i], os.path.join(self.root, shape_names[i], shape_ids[i])) for i
                         in range(len(shape_ids))]
        print(f'The size of {split} data is {len(self.datapath)}')
        self.img_size = img_size
        self.transform = transform
        self._cache = {}
        self.cache_size = cache_size

    def __len__(self):
        return len(self.datapath)

    def __getitem__(self, index):
        if index in self._cache:
            return self._cache[index]

        name, dfile = self.datapath[index]
        cls = self.classes[name]
        cls = np.array([cls]).astype(np.int32)
        xImgFile = dfile + '-X.bin'
        yImgFile = dfile + '-Y.bin'
        zImgFile = dfile + '-Z.bin'

        imgX = np.fromfile(xImgFile, dtype=np.float32).reshape((100,100))
        imgY = np.fromfile(yImgFile, dtype=np.float32).reshape((100,100))
        imgZ = np.fromfile(zImgFile, dtype=np.float32).reshape((100,100))

        imgX[imgX < 1e-5] = 0
        imgY[imgY < 1e-5] = 0
        imgZ[imgZ < 1e-5] = 0

        kcv2.show(imgX, 'x')
        kcv2.show(imgY, 'y')
        kcv2.show_wait(imgZ, 'z')

        img_list = [imgX, imgY, imgZ]
        # random.shuffle(img_list)
        img = cv2.merge(img_list)

        # flip_code = random.randint(-2, 1)
        # if flip_code != -2:
        #     img = cv2.flip(img, flip_code)

        # transpos_code = random.randint(0,1)
        # if transpos_code != 0:
        #     img = cv2.transpose(img)


        if self.img_size[0] != img.shape[0] or self.img_size[1] != img.shape[1]:
            img = cv2.resize(img, self.img_size)

        # kcv2.show_wait(img)
        if self.transform:
            # img = kcv2.cv2pil(img)
            img = self.transform(img)

        if index < self.cache_size:
            self._cache[index] = (img, cls)
        return img, cls


#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/4/24 21:29:16

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import os
import cv2
import numpy as np
from .cvt import keep_bgr, keep_gray, image_size, is_color
#===============================================================================
'''     
'''
#===============================================================================
__all__ = ['read_cv_color', 'read_cv_gray', 'read_pil_img',
           'write_cv_img', 'read_cv_video', 'read_cv_camera',
           'VideoWriter',
           ]

####################    读取图像
def _read_cv_img(file, flags):
    return cv2.imdecode(np.fromfile(file, dtype=np.uint8), flags)

def read_cv_color(file):
    return _read_cv_img(file, cv2.IMREAD_COLOR)

def read_cv_gray(file):
    return _read_cv_img(file, cv2.IMREAD_GRAYSCALE)

def read_pil_img(file):
    from PIL import Image
    return Image.open(file)

def write_cv_img(file, data):
    name, ext = os.path.splitext(file)
    retval, buf = cv2.imencode(ext, data)
    return buf.tofile(file)

####################    读取视频和相机
def _read_cv_capture(fileOrId):
    try:
        cap = cv2.VideoCapture(fileOrId)
        while True:
            ret, frame = cap.read()
            if not ret:
                break
            yield frame
    finally:
        cap.release()

def read_cv_video(video_file):
    for frame in _read_cv_capture(video_file):
        yield frame

def read_cv_camera(camId = 0):
    for frame in _read_cv_capture(camId):
        yield frame


class VideoWriter:
    def __init__(self, filename, fourcc="mp4v", fps=24, frameSize=None, isColor=None):
        self.filename_ = filename
        self.frameSize_ = frameSize
        self.isColor_ = isColor
        self.fps_ = fps
        self.fourcc_ =  cv2.VideoWriter_fourcc(*fourcc)
        self.writer_ = None

    def _make_video(self, img):
        if self.frameSize_ is None:
            self.frameSize_ = image_size(img)

        if self.isColor_ is None:
            self.isColor_ = is_color(img)

        self.writer_ = cv2.VideoWriter(filename=self.filename_, fourcc=self.fourcc_,
                                      fps=self.fps_, frameSize=self.frameSize_, isColor=self.isColor_)

    def write(self, img):
        if self.writer_ is None:
            self._make_video(img)

        if self.isColor_ :
            img = keep_bgr(img)
        else:
            img = keep_gray(img)

        if image_size(img) != self.frameSize_:
            img = cv2.resize(img, dsize=(self.frameSize_[0], self.frameSize_[1]), interpolation=cv2.INTER_LINEAR)

        self.writer_.write(img)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.writer_ is not None:
            self.writer_.release()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/5/26 20:52:29

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''

#===============================================================================
'''     
'''
#===============================================================================
__all__ = ['show_cld_xyz', 'VtkWindow']

class VtkWindow:
    def __init__(self):
        import vtk
        ren = vtk.vtkRenderer()
        ren.SetBackground(0, 0, 0)

        renWin = vtk.vtkRenderWindow()
        renWin.AddRenderer(ren)
        renWin.SetSize(400, 400)
        renWin.SetWindowName('3D viewer')

        iren = vtk.vtkRenderWindowInteractor()
        iren.SetRenderWindow(renWin)
        iren.SetInteractorStyle(vtk.vtkInteractorStyleMultiTouchCamera())

        self.ren_ = ren
        self.renWin_ = renWin
        self.iren_ = iren

    def add_cylinder(self):
        import vtk
        mapper = vtk.vtkPolyDataMapper()
        actor = vtk.vtkActor()
        actor.SetMapper(mapper)

        cylinder = vtk.vtkCylinderSource()
        cylinder.SetResolution(8)
        mapper.SetInputConnection(cylinder.GetOutputPort())

        self.ren_.AddActor(actor)
        return actor

    def add_xyz(self, array, color=(255,255,255)):
        import vtk
        from vtkmodules.util.numpy_support import numpy_to_vtk

        points = vtk.vtkPoints()
        points.SetData(numpy_to_vtk(array))

        polydata = vtk.vtkPolyData()
        polydata.SetPoints(points)

        vertex = vtk.vtkVertexGlyphFilter()
        vertex.SetInputData(polydata)

        mapper = vtk.vtkPolyDataMapper()
        mapper.SetInputConnection(vertex.GetOutputPort())

        actor = vtk.vtkActor()
        actor.SetMapper(mapper)
        actor.GetProperty().SetColor(color[0]/255, color[1]/255, color[2]/255)
        self.ren_.AddActor(actor)
        return actor

    def show(self):
        self.iren_.Initialize()
        self.ren_.ResetCamera()
        self.renWin_.Render()
        self.iren_.Start()


def show_cld_xyz(arr, color=(255,255,255)):
    from ._vtk import VtkWin3d
    win = VtkWin3d()
    win.add_xyz(arr, color=color)
    win.show()




#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/5/20 15:41:47

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import platform, sys
#===============================================================================
'''     
'''
#===============================================================================
__all__ = ['is_linux', 'is_windows', ]


def is_windows():
    return platform.system().lower() == 'windows'

def is_linux():
    return platform.system().lower() == 'linux'

def line_num(depth = 1):
    '''获取当前行号
    '''
    return sys._getframe(depth).f_lineno

def func_name(depth = 1):
    '''获取当前函数名字
    '''
    return sys._getframe(depth).f_code.co_name



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/5/10 17:51:24

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import requests
#===============================================================================
'''     
'''
#===============================================================================

def download_file(url, save_name):
    with open(save_name, 'wb') as f:
        file = requests.get(url)
        f.write(file.content)

def request_text(url):
    return requests.get(url).text

def request_json(url):
    return requests.get(url).json()




#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/5/23 01:32:08

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''

#===============================================================================
'''     
'''
#===============================================================================
__all__ = ['RegressorDecisionTree']

class RegressorDecisionTree:
    def __init__(self,
                 criterion="mse",
                 splitter="best",
                 max_depth=None,
                 min_samples_split=2,
                 min_samples_leaf=1,
                 min_weight_fraction_leaf=0.,
                 max_features=None,
                 random_state=None,
                 max_leaf_nodes=None,
                 min_impurity_decrease=0.,
                 min_impurity_split=None,
                 ccp_alpha=0.0
                 ):

        from sklearn.tree import DecisionTreeRegressor as _Alg
        self._alg = _Alg(criterion=criterion, splitter=splitter, max_depth=max_depth,
                         min_samples_split=min_samples_split, min_samples_leaf=min_samples_leaf,
                         min_weight_fraction_leaf=min_weight_fraction_leaf, max_features=max_features,
                         random_state=random_state, max_leaf_nodes=max_leaf_nodes,
                         min_impurity_decrease=min_impurity_decrease, min_impurity_split=min_impurity_split,
                         ccp_alpha=ccp_alpha
                         )

    def getFeatureImportances(self):
        return self._alg.feature_importances_

    def getMaxFeatures(self):
        return self._alg.max_features

    def getNFeatures(self):
        return self._alg.n_features_

    def getNOutputs(self):
        return self._alg.n_outputs_

    def getTree(self):
        return self._alg.tree_

    def train(self, X, y, sample_weight=None, check_input=True):
        self._alg.fit(X=X, y=y, sample_weight=sample_weight, check_input=check_input)
        return self

    def __call__(self, X, check_input=True):
        return self._alg.predict(X=X, check_input=check_input)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/5/20 21:41:00

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''

#===============================================================================
'''     
'''
#===============================================================================
__all__ = ['iter_multi']

def iter_multi(seq, cnt=1, drop_last=False):
    data_list = []
    for i,v in enumerate(seq):
        data_list.append(v)
        if len(data_list) == cnt:
            yield data_list
            data_list.clear()

    if not drop_last and len(data_list) > 0:
        yield data_list

# def iter_step(seq, step):
#     data_list = []
#     for i,v in enumerate(seq):
#         data_list.append(v)
#         if len(data_list) == cnt:
#             yield data_list
#             data_list.clear()
#
#     if not drop_last and len(data_list) > 0:
#         yield data_list


#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/5/20 21:01:14

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''

#===============================================================================
'''     
'''
#===============================================================================

__all__ = ['file_read_text', 'file_read_bin', 'file_write_text', 'file_write_bin',
           'file_read_lines']

def file_read_text(fname, encoding=None):
    with open(fname, mode='r', encoding=encoding) as f:
        return f.read()

def file_read_lines(fname, encoding=None, strip = True):
    with open(fname, mode='r', encoding=encoding) as f:
        line = f.readline()
        if strip:
            line = line.strip()
        yield line

def file_read_bin(fname):
    with open(fname, mode='rb') as f:
        return f.read()

def file_write_text(fname, text, encoding=None):
    with open(fname, mode='w', encoding=encoding) as f:
        f.write(text)

def file_write_bin(fname, data):
    with open(fname, mode='wb') as f:
        f.write(data)




#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/4/20 20:59:02

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import torch
#===============================================================================
'''     
'''
#===============================================================================
def print2(data, desc = None):
    '''show more data detail
    :param data:
    :return:
    '''
    if torch.is_tensor(data):
        if desc:
            print(f'===== torch.Tensor {desc}')
        else:
            print(f'===== torch.Tensor')
        print(f'type:{data.type()}')
        print(f'size:{data.size()}')
        print(f'dim:{data.dim()}')
        print(data)
    else:
        print(data)

def print_version():
    print(f'torch.__version__:{torch.__version__}')
    print(f'torch.version.cuda:{torch.version.cuda}')
    print(f'torch.backends.cudnn.version():{torch.backends.cudnn.version()}')
    print(f'torch.cuda.get_device_name():{torch.cuda.get_device_name()}')

def print_info():
    print_version()
    print(f'torch.get_default_dtype():{torch.get_default_dtype()}')

    print(f'torch.compiled_with_cxx11_abi():{torch.compiled_with_cxx11_abi()}')
    print(f'torch.is_warn_always_enabled():{torch.is_warn_always_enabled()}')
    print(f'torch.are_deterministic_algorithms_enabled():{torch.are_deterministic_algorithms_enabled()}')
    print(f'torch.is_deterministic_algorithms_warn_only_enabled():{torch.is_deterministic_algorithms_warn_only_enabled()}')


def get_device(idx = 0):
    '''get the suitable device
    :param idx:
    :return:
    '''
    return torch.device(f"cuda:{idx}" if torch.cuda.is_available() else "cpu")

def init_seeds(seed=0):
    '''make the torch enviroment no random.
    '''

    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False
    # # Speed-reproducibility tradeoff https://pytorch.org/docs/stable/notes/randomness.html
    # if seed == 0:  # slower, more reproducible
    #     torch.cudnn.benchmark, torch.cudnn.deterministic = False, True
    # else:  # faster, less reproducible
    #     torch.cudnn.benchmark, torch.cudnn.deterministic = True, False

def cuda_visible_devices(device : str):
    ''' input device str list like: '0,1'
    or usage in command:
        CUDA_VISIBLE_DEVICES=0,1 python train.py
    :param device:
    :return:
    '''
    import os
    os.environ['CUDA_VISIBLE_DEVICES'] = str(device)




#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/5/20 08:58:19

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import os
import torch
from .base import save_model_params, load_model_params
from laok.util.time import datetime_str
#===============================================================================
'''     
'''
#===============================================================================
__all__ = ['ModelCheckPoint']

def _split_file_name(filename):
    try:
        fname = os.path.split(filename)[1]
        basename = os.path.splitext(fname)[0]
        value = basename.split('-')[-1]
        value = float(value)
    except Exception as e:
        value = None

    return fname, value

def _save_file_name(model_path, name, value):
    dt = datetime_str()
    _save_file = os.path.join(model_path, f'{name}-{dt}-{value:.4f}.pt')
    return _save_file

class ModelCheckPoint:
    '''automatically load or save the current best model
    '''
    def __init__(self, model, model_save_path='./model_save/', name='model', keep_best=True, show_info=True):
        self.model_ = model
        self.model_save_path_ = model_save_path
        self.name_ = name
        self.value_ = None
        self.show_info_ = show_info
        self.keep_best_ = keep_best
        if not os.path.exists(model_save_path):
            os.makedirs(model_save_path)

    def update(self, value):
        '''the model needs to be saved according to the value
        '''
        _save_file = None

        # check if need save model
        if self.value_ is None or self.value_ < value:
            self.value_ = value  # update the best value
            _save_file = _save_file_name(self.model_save_path_, name=self.name_, value = value)
            save_model_params(self.model_, _save_file)
            if self.show_info_:
                print(f'save checkpoint {_save_file}')

        if self.show_info_:
            if self.value_ == value:
                print(f"current {self.name_} value is {value:.4f}, it's the best.!!!")
            else:
                print(f'current {self.name_} value is {value:.4f}, while the best value is {self.value_:.4f}')

        # load the best one model from model path
        if self.keep_best_:
            self.load_from_path()

        return _save_file

    def load_from_path(self, ext=".pt"):
        max_value = None
        max_value_file = None

        for fname in os.listdir(self.model_save_path_):
            if not fname.endswith(ext):
                continue

            _fname, _value = _split_file_name(fname)

            # print(_fname, _value)
            if max_value is None or max_value < _value:
                max_value = _value
                max_value_file = _fname

        if max_value_file and max_value and (self.value_ is None or max_value > self.value_):
            self.value_ = max_value
            _load_file = os.path.join(self.model_save_path_, max_value_file)
            load_model_params(_load_file, self.model_)
            if self.show_info_:
                # if _dist.is_initialized():
                #     print(f'rank[{_dist.get_rank()}], load checkpoint: {_load_file}, value:{self._value}')
                # else:
                print(f'load checkpoint: {_load_file}, value:{self.value_}')

        return max_value, max_value_file

    def load_from_file(self, model_file, value=None):
        _fname, _value = _split_file_name(model_file)
        _dir = os.path.split(model_file)[0]

        if os.path.abspath(_dir) == os.path.abspath(self.model_save_path_):
            self.value_ = _value if value is None else value

        load_model_params(model_file, self.model_)
        if self.show_info_:
            # if _dist.is_initialized():
            #     print(f'rank[{_dist.get_rank()}], load model file: {model_file}, value:{self._value}')
            # else:
            print(f'load model file: {model_file}, value:{self.value_}')



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/5/20 12:16:18

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import numpy as np
import random
#===============================================================================
'''     
'''
#===============================================================================

__all__ = ['print2', 'init_seeds']

def print2(data, name = None):
    '''print detail data information
    '''
    msg = f"==== name:{name} " if name else "=== "
    msg += f'id:{id(data)} type:{type(data)}\n'

    if isinstance(data, np.ndarray):
        msg += f'    shape:{data.shape}\n' \
               f'    dtype:{data.dtype}\n'

    if isinstance(data, np.dtype):
        if data.names:
            msg += f'    names:{data.names}\n'

        if data.shape:
            msg += f'    shape:{data.shape}\n'
            msg += f'    dim:{data.dim}\n'

        if data.hasobject:
            msg += f'    hasobject:{data.hasobject}\n'


        msg += f'    itemsize:{data.itemsize}\n' \
               f'    alignment:{data.alignment}\n' \
               f'    str:{data.str}\n' \
               f'    byteorder:{data.byteorder}\n'

    msg += f'{data}'
    print(msg)


def init_seeds(seed=0):
    # Initialize random number generator (RNG) seeds
    random.seed(seed)
    np.random.seed(seed)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/5/12 21:33:43

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import matplotlib.pyplot as plt
#===============================================================================
'''     
'''
#===============================================================================
__all__ = ['draw_bar', 'draw_fill_between', 'draw_plot', 'draw_scatter', 'draw_stem', 'draw_step',
            'draw_boxplot', 'draw_errorbar', 'draw_eventplot', 'draw_hexbin', 'draw_hist',
            'draw_hist2d', 'draw_pie', 'draw_violinplot', 'draw_barbs', 'draw_contour',
            'draw_contourf', 'draw_imshow', 'draw_pcolormesh', 'draw_quiver', 'draw_streamplot',
            'draw_tricontour', 'draw_tricontourf', 'draw_tripcolor', 'draw_triplot'
           ]
################### basic

def _get_kws(d):
    d = d.copy()
    d.pop('kwargs')
    for k in list(d.keys()):
        if d[k] is None:
            d.pop(k)
    return d

def draw_bar(x, height, width=0.8, bottom=None, align="center",
             color=None, edgecolor=None, linewidth=None, tick_label=None,
             xerr=None, yerr=None, ecolor=None, capsize=None, error_kw=None, log=None,
             **kwargs):
    plt.bar(**_get_kws(locals()))

def draw_barh(*args, **kwargs):
    return plt.barh(*args, **kwargs)

def draw_fill_between(*args, **kwargs):
    return plt.fill_between(*args, **kwargs)

def draw_plot(*args, **kwargs):
    return plt.plot(*args, **kwargs)

def draw_scatter(*args, **kwargs):
    return plt.scatter(*args, **kwargs)

def draw_stem(*args, **kwargs):
    return plt.stem(*args, **kwargs)

def draw_step(*args, **kwargs):
    return plt.step(*args, **kwargs)

################### stats
def draw_boxplot(*args, **kwargs):
    return plt.boxplot(*args, **kwargs)

def draw_errorbar(*args, **kwargs):
    return plt.errorbar(*args, **kwargs)

def draw_eventplot(*args, **kwargs):
    return plt.eventplot(*args, **kwargs)

def draw_hexbin(*args, **kwargs):
    return plt.hexbin(*args, **kwargs)

def draw_hist(*args, **kwargs):
    return plt.hist(*args, **kwargs)

def draw_hist2d(*args, **kwargs):
    return plt.hist2d(*args, **kwargs)

def draw_pie(*args, **kwargs):
    return plt.pie(*args, **kwargs)

def draw_violinplot(*args, **kwargs):
    return plt.violinplot(*args, **kwargs)

################### arrays
def draw_barbs(X,Y,U,V, pivot='tip', length=7, barbcolor=None, flagcolor=None,
                 sizes=None, fill_empty=False, barb_increments=None,
                 rounding=True, flip_barb=False, **kws):
    plt.barbs(X,Y,U,V, pivot=pivot, length=length, barbcolor=barbcolor, flagcolor=flagcolor,
                sizes=sizes, fill_empty=fill_empty, barb_increments=barb_increments,
                rounding=rounding, flip_barb=flip_barb, **kws)

def draw_contour(*args, **kwargs):
    return plt.contour(*args, **kwargs)

def draw_contourf(*args, **kwargs):
    return plt.contourf(*args, **kwargs)

def draw_imshow(*args, **kwargs):
    return plt.imshow(*args, **kwargs)

def draw_pcolormesh(*args, **kwargs):
    return plt.pcolormesh(*args, **kwargs)

def draw_quiver(*args, **kwargs):
    return plt.quiver(*args, **kwargs)

def draw_streamplot(*args, **kwargs):
    return plt.streamplot(*args, **kwargs)

################### unstructured

def draw_tricontour(*args, **kwargs):
    return plt.tricontour(*args, **kwargs)

def draw_tricontourf(*args, **kwargs):
    return plt.tricontourf(*args, **kwargs)

def draw_tripcolor(*args, **kwargs):
    return plt.tripcolor(*args, **kwargs)

def draw_triplot(*args, **kwargs):
    return plt.triplot(*args, **kwargs)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/5/10 14:19:19

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import cv2
#===============================================================================
'''     
'''
#===============================================================================
__all__ = ['draw_text']

def draw_rect():
    pass

def draw_line():
    pass

def draw_circle():
    pass

def draw_ellipse():
    pass

def draw_polygon():
    pass

def draw_contours():
    pass

def draw_point():
    pass

def draw_detect_rect():
    pass

def draw_text(img, text, org=(100,100), fontFace=cv2.FONT_HERSHEY_SIMPLEX, fontScale=1.0, color=(0,0,255), thickness=1, lineType=cv2.LINE_AA):
    return cv2.putText(img, text, org = org, fontFace=fontFace, fontScale = fontScale,
                color=color, thickness=thickness, lineType=lineType, bottomLeftOrigin=False)


#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/3/7 15:43:22

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import os
import numpy as np
from torch.utils.data import Dataset
import laok.cv3d as kcv3
import laok.cv2d as kcv2
import cv2
#===============================================================================
# 
#===============================================================================
__all__ = ['ModelNetProjImage']


class ModelNetProjImage(Dataset):
    def __init__(self, root, div=500, split='train', num_category=40, cache_size=30000, transform=None):
        assert (split == 'train' or split == 'test')

        # 加载类别列表
        catfile = os.path.join(root, f'modelnet{num_category}_shape_names.txt')
        with open(catfile) as f:
            self.classes = {line.rstrip(): i for i, line in enumerate(f)}
        print(f'classes : {self.classes}')

        # 加载数据文件列表
        shape_name_file = f'modelnet{num_category}_{split}.txt'
        print(f'{split} file : {shape_name_file}')
        with open(os.path.join(root, shape_name_file))as f:
            shape_ids = [line.rstrip() for line in f]
        shape_names = ['_'.join(x.split('_')[0:-1]) for x in shape_ids]  # 去除尾数
        # list of (shape_name, shape_txt_file_path) tuple
        self.datapath = [(shape_names[i], os.path.join(root, shape_names[i], shape_ids[i])) for i
                         in range(len(shape_ids))]
        print(f'The size of {split} data is {len(self.datapath)}')
        self.div = div
        self._cache = {}
        self.cache_size = cache_size
        self.transform = transform

    def __len__(self):
        return len(self.datapath)

    def __getitem__(self, index):
        if index in self._cache:
            return self._cache[index]

        name, dfile = self.datapath[index]
        cls = self.classes[name]
        cls = np.array([cls]).astype(np.int32)

        cld = kcv3.load_cld_xyz(dfile + ".txt", delimiter=',')
        img_x = kcv3.proj_img_x_div(cld, self.div)
        img_y = kcv3.proj_img_y_div(cld, self.div)
        img_z = kcv3.proj_img_z_div(cld, self.div)

        img_x = kcv2.normlize_uint8(img_x)
        img_y = kcv2.normlize_uint8(img_y)
        img_z = kcv2.normlize_uint8(img_z)

        img_list = [img_x, img_y, img_z]
        img = cv2.merge(img_list)

        if self.transform:
            img = self.transform(img)

        if index < self.cache_size:
            self._cache[index] = (img, cls)
        return img, cls


#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/4/14 16:47:51

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
from collections import OrderedDict
#===============================================================================
'''     
'''
#===============================================================================
__all__ = ['ClsIdx']

class ClsIdx:
    def __init__(self, cls_file = None):
        self.cls_data_ = OrderedDict()
        if cls_file:
            self.setClsFile(cls_file)

    def setClsFile(self, cls_file):
        with open(cls_file) as f:
            for i,line in enumerate(f):
                self.cls_data_[i] = line.strip()
        return self

    def getName(self, idx):
        _idx = int(idx)
        return self.cls_data_.get(_idx, '')

    def getIndex(self, name):
        if name == None or name == '':
            return -1
        for _idx, _name in self.cls_data_.items():
            if name in _name:
                return _idx
        return -1

    def show(self):
        for k,v in self.cls_data_.items():
            print(f'{k}={v}')



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/5/12 21:23:44

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import matplotlib.pyplot as plt
import numpy as np
#===============================================================================
'''     
'''
#===============================================================================

__all__ = ['draw_data_clusters',
           ]

def draw_data_clusters(X, labels, index_list=(0, 1)):
    clusters = np.unique(labels)
    x, y = index_list
    for cluster in clusters:
        row_ix = np.where(labels == cluster)
        plt.scatter(X[row_ix, x], X[row_ix, y])




#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/4/19 11:14:26

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import onnx
import onnxruntime
import laok.util.fpath as kpath
#===============================================================================
'''
'''
#===============================================================================

def check_model(onnxFile):
    '''check onnx model if it is valide
    :param onnxFile:
    :return:
    '''
    model = onnx.load_model(onnxFile)
    onnx.checker.check_model(model)

def load_model(onnxFile):
    '''
    :param onnxFile:
    :return:
    '''
    return onnx.load_model(onnxFile)


def __fmt_NodeArg(nodes):
    msg_list = []
    for node in nodes:
        shape_name = "(" + ", ".join( str(s) for s in node.shape) + ")"
        msg_list.append(f'    {node.name} : {shape_name} , type={node.type}')
    return msg_list

def dump_model_info(onnxFile, graphFile = None, skip_exist = False, show_info = True):
    '''
    :param onnxFile:
    :param graphFile:
    :param skip_exist:
    :param show_info:
    :return:
    '''
    if graphFile is None:
        graphFile = kpath.path_replace_ext(onnxFile, '.graph')

    if skip_exist and kpath.path_exist(graphFile):
        return graphFile

    if show_info:
        print(f'start read... {onnxFile}')

    msg_list = []
    ###################### onnxruntime的模型信息
    ort_session = onnxruntime.InferenceSession(onnxFile)
    # 模型基本信息
    meta = ort_session.get_modelmeta()
    msg_list.append('modelMeta:')
    msg_list.append(f'    producer_name:{meta.producer_name}')
    msg_list.append(f'    graph_name:{meta.graph_name}')
    msg_list.append(f'    domain:{meta.domain}')
    msg_list.append(f'    description:{meta.description}')
    msg_list.append(f'    graph_description:{meta.graph_description}')
    msg_list.append(f'    version:{meta.version}')

    # 获取输入信息
    inputs = ort_session.get_inputs()
    msg_list.append(f'\n\nInput Node Name/Shape [{len(inputs)}]:')
    msg_list.extend(__fmt_NodeArg(inputs))

    # 获取输出信息
    outputs = ort_session.get_outputs()
    msg_list.append(f'\n\nOutput Node Name/Shape [{len(outputs)}]:')
    msg_list.extend(__fmt_NodeArg(outputs))

    # 获取重载初始化
    overinits = ort_session.get_overridable_initializers()
    msg_list.append(f'\n\nOverridableInitializer Node Name/Shape [{len(overinits)}]:')
    msg_list.extend(__fmt_NodeArg(overinits))


    ###################### onnx 的模型信息
    try:
        model = onnx.load_model(onnxFile)
        graph = onnx.helper.printable_graph(model.graph)
        msg_list.append("\n\n" + graph)
    except Exception as e:
        print(e)
    ###################### 保存onnx模型信息
    with open(graphFile, 'w') as f:
        f.write('\n'.join(msg_list))

    if show_info:
        print(f'dump file... {onnxFile}')

    return graphFile



from .cvt import *
from .draw import *
from .gui import *
from .io import *


from .aug import *
from .model import *
from .edge import *
from .filter import *
from .morph import *
from .misc import *
from .photo import *
from .trans import *



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/4/21 21:28:54

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import torch
import torch.nn as nn
#===============================================================================
'''         
'''
#===============================================================================


def train_class_model(model, train_loader, optimizer, num_epochs):
    # Loss and optimizer
    criterion = nn.CrossEntropyLoss()
    optimizer = torch.optim.Adam(model.parameters(), lr=learning_rate)

    # Train the model
    total_step = len(train_loader)
    for epoch in range(num_epochs):
        for i ,(images, labels) in enumerate(train_loader):
            images = images.to(device)
            labels = labels.to(device)

            # Forward pass
            outputs = model(images)
            loss = criterion(outputs, labels)

            # Backward and optimizer
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

            if (i+1) % 100 == 0:
                print(f'Epoch: [{epoch+1}/{num_epochs}], Step: [{i+1}/{total_step}], Loss: {loss.item()}')


def val_class_model(model, loader, device=None):
    from . import get_device
    if device is None:
        device = get_device()
    from tqdm import tqdm
    model.eval()  # eval mode(batch norm uses moving mean/variance instead of mini-batch mean/variance)
    correct = 0
    total = 0
    img_total = 0
    with torch.no_grad():
        for images, labels in loader:
            images = images.to(device)
            labels = labels.to(device)
            img_total += len(images)
            outputs = model(images)
            _, predicted = torch.max(outputs.data, 1)
            total += labels.size(0)
            correct += (predicted == labels).sum().item()
        print(f'accuracy of the model on the {img_total} test images: {100 * correct / total} %')


#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/5/6 11:06:36

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import cv2
import numpy as np
#===============================================================================
'''  
# skimage 图像格式        通道：RGB 像素值：[0.0,1.0]      (h,w)
    
# Opencv  图像格式        通道：BGR 像素值：[0,255]        (h,w)
    gray是 (h,w)  
    color是(h,w,3)
# PIL     图像格式        (w,h)   
    
'''
#===============================================================================

__all__  = ['gray2bgr', 'gray2rgb',
            'bgr2gray', 'rgb2gray',
            'swap_rb',
            'cv2pil', 'pil2cv', 'cv2sk', 'sk2cv', 'pil2sk', 'sk2pil',
            'normlize_uint8', 'normlize_float32',
            'is_color', 'is_gray',
            'keep_bgr', 'keep_gray',
            'image_size', 'resize_abs', 'resize_rel'
            ]

def gray2bgr(img):
    if img.ndim == 2:
        img = cv2.cvtColor(img, cv2.COLOR_GRAY2BGR)
    return img

def gray2rgb(img):
    if img.ndim == 2:
        img = cv2.cvtColor(img, cv2.COLOR_GRAY2RGB)
    return img

def bgr2gray(img):
    if img.ndim == 3:
        if img.shape[-1] == 3:
            img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        elif img.shape[-1] == 4:
            img = cv2.cvtColor(img, cv2.COLOR_BGRA2GRAY)
    return img

def rgb2gray(img):
    if img.ndim == 3:
        if img.shape[-1] == 3:
            img = cv2.cvtColor(img, cv2.COLOR_RGB2GRAY )
        elif img.shape[-1] == 4:
            img = cv2.cvtColor(img, cv2.COLOR_RGBA2GRAY )
    return img

def swap_rb(img):
    if img.ndim == 3:
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
    return img

def cv2pil(img):
    from PIL import Image
    if img.ndim == 3 and img.shape[-1] == 3:
        img = Image.fromarray(cv2.cvtColor(img,cv2.COLOR_BGR2RGB), mode = 'RGB')
    else:
        img = Image.fromarray(img, mode = 'L')
    return img

def pil2cv(img):
    if img.mode == 'RGB':
        image = cv2.cvtColor(np.asarray(img), cv2.COLOR_RGB2BGR)
    elif img.mode == 'RGBA':
        image = cv2.cvtColor(np.asarray(img), cv2.COLOR_RGBA2BGRA)
    elif img.mode == 'L':
        image = np.asarray(img)
    return image

def cv2sk(img):
    if img.ndim == 3:
        if img.shape[-1] == 3:
            img = cv2.cvtColor(img,cv2.COLOR_BGR2RGB)
        elif img.shape[-1] == 4:
            img = cv2.cvtColor(img,cv2.COLOR_BGRA2RGBA)
    return img

def sk2cv(img):
    if img.ndim == 3:
        if img.shape[-1] == 3:
            img = cv2.cvtColor(img,cv2.COLOR_RGB2BGR)
        elif img.shape[-1] == 4:
            img = cv2.cvtColor(img,cv2.COLOR_RGBA2BGRA)
    return img

def pil2sk(img):
    pass

def sk2pil(img):
    pass

def is_color(img):
    if img.ndim == 3:
        return img.shape[-1] > 1
    return False

def is_gray(img):
    return not is_color(img)

def normlize_uint8(mat, alpha=0, beta=255):
    return cv2.normalize(mat, dst=None, alpha=alpha, beta=beta, norm_type=cv2.NORM_MINMAX, dtype=cv2.CV_8U)

def normlize_float32(mat, alpha=0, beta=1):
    return cv2.normalize(mat, dst=None, alpha=alpha, beta=beta, norm_type=cv2.NORM_MINMAX, dtype=cv2.CV_32F)

def normlize_int32(mat, alpha=0, beta=255):
    return cv2.normalize(mat, dst=None, alpha=alpha, beta=beta, norm_type=cv2.NORM_MINMAX, dtype=cv2.CV_32S)

def keep_gray(img):
    if is_color(img):
        return bgr2gray(img)
    return img

def keep_bgr(img):
    if is_gray(img):
        return gray2bgr(img)
    return img

def image_size(img):
    return (img.shape[1], img.shape[0])

def resize_abs(img, width, height, interpolation=cv2.INTER_LINEAR):
    return cv2.resize(img, dsize=(int(width), int(height)), interpolation=interpolation)

def resize_rel(img, fx= 0.5, fy = 0, interpolation=cv2.INTER_LINEAR):
    if fy == 0:
        fy = fx
    return cv2.resize(img, fx=fx, fy=fy, interpolation=interpolation)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2021/3/17 21:11:08

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import time as _time
import sys
from contextlib import contextmanager as _context
from functools import wraps as _wraps
from collections import OrderedDict as _OrderedDict
from datetime import datetime
#===============================================================================
# 
#===============================================================================
__all__ = ['datetime_str', 'auto_timer', 'Timer',  'deco_time', 'TimerLabel']

def datetime_str(fmt='%Y%m%d%H%M%S', dt = None):
    if dt is None:
        dt = datetime.now()
    return dt.strftime('%Y%m%d%H%M%S')

@_context
def auto_timer(msg = '', need_print = True):
    t1 = _time.time()
    yield
    t2 = _time.time()
    if need_print:
        print(f'{msg} [time:{(t2-t1)*1000}(ms)]')

def deco_time(stream = None):
    '''
    装饰器,记录时间
    '''
    def _w1(func):
        @_wraps(func)
        def _w2(*args, **kwargs):
            start = _time.time()
            res = func(*args, **kwargs)
            stop = _time.time()
            if _time is None:
                stream = sys.stdout
            stream.write(f"[{ func.__name__ }] use time [{ 1000*(stop-start) }(ms)]\n")
            return res
        return _w2
    return _w1

class Timer:
    def __init__(self):
        self.restart()

    def restart(self):
        self._t1 = _time.time()

    def elapse(self):
        '''
        time in seconds since the Epoch
        :return:
        '''
        return _time.time() - self._t1


class TimerLabel:
    def __init__(self):
        self._t1 = _OrderedDict()
        self._t2 = _OrderedDict()
        self._enabled = True
        self._total_count = None
        self._cur_count = 0

    def setEnabled(self, val):
        self._enabled = val

    def setReportCount(self, cnt):
        self._total_count = cnt
        self._cur_count = 0

    def start(self, name):
        if self._enabled:
            self._t1[name] = _time.time()

    def end(self, name):
        if self._enabled:
            self._t2[name] = _time.time()            

    def report(self, stream = None):
        if not self._enabled:
            return 
        
        if self._total_count is not None:
            if self._cur_count > self._total_count:
                return
            self._cur_count += 1

        if stream is None:
            stream = sys.stdout
        
        curt = _time.time()
        stream.write("\n***** time summary *****\n")
        for name,t1 in self._t1.items():
            t2 = self._t2.get(name, curt)
            stream.write(f'"{name}" time use: {t2-t1}(s)\n')


if __name__ == "__main__":
    import math
    a = 0
    with auto_timer() as t:
        for i in range(1,10000000):
            a += math.cos(1)

    times = TimerLabel()
    times.setReportCount(2)

    for i in range(5):
        times.start('it1')
        _time.sleep(1)
        times.start('it2')
        _time.sleep(3)
        times.report()


#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/5/26 10:16:16

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import cv2
#===============================================================================
'''     
'''
#===============================================================================
__all__ = ['morph_close', 'morph_open', 'morph_blackhat',
           'morph_dilate', 'morph_erode', 'morph_gradient',
           'morph_hitmiss', 'morph_tophat']

def _morph_impl(img, op, ksize=(3, 3), shape=cv2.MORPH_RECT, anchor=(-1, -1), iterations=1, borderType=cv2.BORDER_CONSTANT, borderValue=(0, 0, 0)):
    kernel = cv2.getStructuringElement(shape=shape, ksize=ksize, anchor=anchor)
    return cv2.morphologyEx(img, op=op, kernel=kernel, anchor=anchor, iterations=iterations, borderType=borderType, borderValue=borderValue)

def morph_close(img, ksize=(3,3), shape=cv2.MORPH_RECT, anchor=(-1,-1), iterations=1, borderType=cv2.BORDER_CONSTANT, borderValue=(0,0,0)):
    return _morph_impl(img, op=cv2.MORPH_CLOSE, ksize=ksize, shape=shape, anchor=anchor, iterations=iterations, borderType=borderType, borderValue=borderValue)

def morph_open(img, ksize=(3,3), shape=cv2.MORPH_RECT, anchor=(-1,-1), iterations=1, borderType=cv2.BORDER_CONSTANT, borderValue=(0,0,0)):
    return _morph_impl(img, op=cv2.MORPH_OPEN, ksize=ksize, shape=shape, anchor=anchor, iterations=iterations, borderType=borderType, borderValue=borderValue)

def morph_dilate(img, ksize=(3,3), shape=cv2.MORPH_RECT, anchor=(-1,-1), iterations=1, borderType=cv2.BORDER_CONSTANT, borderValue=(0,0,0)):
    return _morph_impl(img, op=cv2.MORPH_DILATE, ksize=ksize, shape=shape, anchor=anchor, iterations=iterations, borderType=borderType, borderValue=borderValue)

def morph_erode(img, ksize=(3,3), shape=cv2.MORPH_RECT, anchor=(-1,-1), iterations=1, borderType=cv2.BORDER_CONSTANT, borderValue=(0,0,0)):
    return _morph_impl(img, op=cv2.MORPH_ERODE, ksize=ksize, shape=shape, anchor=anchor, iterations=iterations, borderType=borderType, borderValue=borderValue)

def morph_gradient(img, ksize=(3,3), shape=cv2.MORPH_RECT, anchor=(-1,-1), iterations=1, borderType=cv2.BORDER_CONSTANT, borderValue=(0,0,0)):
    return _morph_impl(img, op=cv2.MORPH_GRADIENT, ksize=ksize, shape=shape, anchor=anchor, iterations=iterations, borderType=borderType, borderValue=borderValue)

def morph_tophat(img, ksize=(3,3), shape=cv2.MORPH_RECT, anchor=(-1,-1), iterations=1, borderType=cv2.BORDER_CONSTANT, borderValue=(0,0,0)):
    return _morph_impl(img, op=cv2.MORPH_TOPHAT, ksize=ksize, shape=shape, anchor=anchor, iterations=iterations, borderType=borderType, borderValue=borderValue)

def morph_blackhat(img, ksize=(3,3), shape=cv2.MORPH_RECT, anchor=(-1,-1), iterations=1, borderType=cv2.BORDER_CONSTANT, borderValue=(0,0,0)):
    return _morph_impl(img, op=cv2.MORPH_BLACKHAT, ksize=ksize, shape=shape, anchor=anchor, iterations=iterations, borderType=borderType, borderValue=borderValue)

def morph_hitmiss(img, ksize=(3,3), shape=cv2.MORPH_RECT, anchor=(-1,-1), iterations=1, borderType=cv2.BORDER_CONSTANT, borderValue=(0,0,0)):
    return _morph_impl(img, op=cv2.MORPH_HITMISS, ksize=ksize, shape=shape, anchor=anchor, iterations=iterations, borderType=borderType, borderValue=borderValue)




#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/5/20 08:58:48

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import torch
#===============================================================================
'''     
'''
#===============================================================================
__all__ = ['save_jit_model', 'save_onnx_model', 'save_model', 'save_model_params',
           'load_model_params', 'load_jit_model'
           ]

def save_jit_model(model, filename, input = None):
    '''save trace model,use eval model to get trace model;
    '''
    if isinstance(model, torch.jit.TracedModule):
        traced_model = model
    else:
        try:
            mode = model.training
            model.eval()
            traced_model = torch.jit.trace(model, input)
        finally:
            model.train(mode)

    traced_model.save(filename)
    return traced_model

def load_jit_model(filename):
    ''' load trace model
    '''
    return torch.jit.load(filename)

def save_onnx_model(model, filename, input = None, **kws):
    '''export model to onnx
    :param model:
    :param filename:
    :param input:
    :return:
    '''
    torch.onnx.export(model, input, filename, **kws)

def save_model(model, filename):
    return torch.save(model, filename)

def save_model_params(model, filename):
    return torch.save(model.state_dict(), filename)

def load_model(filename, map_location=None):
    return torch.load(filename, map_location=map_location)

def load_model_params(filename, model, map_location=None):
    _obj = load_model(filename, map_location=map_location)
    if isinstance(_obj, torch.nn.Module):
        state = _obj.state_dict()
    else:
        state = _obj

    # 加载匹配的部分
    _model_state = model.state_dict()
    part_load = {}
    for k,v in state.items():
        if k in _model_state and _model_state[k].shape == v.shape:
            part_load[k] = v
    _model_state.update(part_load)
    model.load_state_dict(_model_state)
    return model





#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/4/29 17:22:41

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import cv2, numpy as np
#===============================================================================
'''     
'''
#===============================================================================

__all__ = ['show', 'show_wait', 'wait']

def show(mat, winname='img', winflag = cv2.WINDOW_NORMAL, title = None):
    cv2.namedWindow(winname, winflag)
    if title:
        cv2.setWindowTitle(winname, title)
    if mat.dtype == np.bool: #转换bool 类型
        mat = mat.astype(np.uint8) * 255
    elif mat.dtype != np.uint8:  # 转换 其它类型
        mat = cv2.normalize(mat, None, alpha=0, beta=255, norm_type=cv2.NORM_MINMAX, dtype=cv2.CV_8U)
    cv2.imshow(winname, mat)

def wait(delay=0):
    key = cv2.waitKey(delay)
    if key == 27: # 'Esc' to quit
        quit()

def show_wait(mat, winname='img', delay=0, winflag = cv2.WINDOW_NORMAL, title = None):
    show(mat, winname=winname, winflag=winflag, title=title)
    wait(delay=delay)



# -*- coding:utf-8 -*-
# author: Xinge

"""
SemKITTI dataloader
"""
import os
import numpy as np
import torch
import random
import time
import numba as nb
from torch.utils import data
from pathlib import Path
import pickle
#===============================================================================
#
#===============================================================================
__all__ = ['SemanticKittiSplitData',  'SemanticKittiData', 'SemanticKittiVoxel', 'SemanticKittiCylinder', 'SemanticKittiPolar', 'collate_fn_BEV']

def absoluteFilePaths(directory):
    for dirpath, _, filenames in os.walk(directory):
        for f in filenames:
            yield os.path.abspath(os.path.join(dirpath, f))

class SemanticKittiSplitData(data.Dataset):
    def __init__(self, data_file, start=0, stop=1, imageset='train'):
        import yaml
        label_mapping = Path(__file__).parent.joinpath("semantic-kitti.yaml")

        # print(f'open label mapping:{label_mapping}')
        with open(label_mapping, 'r') as stream:
            semkittiyaml = yaml.safe_load(stream)

        self.semkittiyaml = semkittiyaml
        self.learning_map = semkittiyaml['learning_map']

        import json
        with open(data_file, 'r') as f:
            jdata = json.load(f)
        dir_name = Path(data_file).parent
        self.im_idx = []

        for data in jdata[start:stop]:
            self.im_idx.extend( str(dir_name.joinpath(v)) for v in data)

        for binfile in self.im_idx:
            print(binfile)
        print(f'len data = {len(self.im_idx)}')

    def __len__(self):
        'Denotes the total number of samples'
        return len(self.im_idx)

    def __getitem__(self, index):
        bin_file = self.im_idx[index]
        # print(f"bin file: {bin_file}")
        raw_data = np.fromfile(bin_file, dtype=np.float32).reshape((-1, 4))
        label_file = self.im_idx[index].replace('velodyne', 'labels')[:-3] + 'label'
        # print(f"label file: {label_file}")
        annotated_data = np.fromfile(label_file, dtype=np.int32).reshape((-1, 1))
        annotated_data = annotated_data & 0xFFFF  # delete high 16 digits binary
        annotated_data = np.vectorize(self.learning_map.__getitem__)(annotated_data)
        return raw_data[:, :3], annotated_data.astype(np.uint8),  raw_data[:, 3]


class SemanticKittiData(data.Dataset):
    def __init__(self, data_path, imageset='train'):
        import yaml
        label_mapping = Path(__file__).parent.joinpath("semantic-kitti.yaml")

        # print(f'open label mapping:{label_mapping}')
        with open(label_mapping, 'r') as stream:
            semkittiyaml = yaml.safe_load(stream)

        self.semkittiyaml = semkittiyaml
        self.learning_map = semkittiyaml['learning_map']
        self.imageset = imageset
        if imageset == 'train':
            split = semkittiyaml['split']['train']
        elif imageset == 'val':
            split = semkittiyaml['split']['valid']
        elif imageset == 'test':
            split = semkittiyaml['split']['test']
        else:
            raise Exception('Split must be train/val/test')

        self.im_idx = []
        for i_folder in split:
            self.im_idx += absoluteFilePaths('/'.join([data_path, str(i_folder).zfill(2), 'velodyne']))
            break
        # for binfile in self.im_idx:
        #     print(binfile)

    def __len__(self):
        'Denotes the total number of samples'
        return len(self.im_idx)

    def __getitem__(self, index):
        bin_file = self.im_idx[index]
        # print(f"bin file: {bin_file}")
        raw_data = np.fromfile(bin_file, dtype=np.float32).reshape((-1, 4))
        if self.imageset == 'test':
            annotated_data = np.expand_dims(np.zeros_like(raw_data[:, 0], dtype=int), axis=1)
        else:
            label_file = self.im_idx[index].replace('velodyne', 'labels')[:-3] + 'label'
            # print(f"label file: {label_file}")
            annotated_data = np.fromfile(label_file, dtype=np.int32).reshape((-1, 1))
            annotated_data = annotated_data & 0xFFFF  # delete high 16 digits binary
            annotated_data = np.vectorize(self.learning_map.__getitem__)(annotated_data)

        return raw_data[:, :3], annotated_data.astype(np.uint8),  raw_data[:, 3]


class SemanticKittiVoxel(data.Dataset):
    def __init__(self, in_dataset, grid_size = (480, 360, 32), rotate_aug=False, flip_aug=False, ignore_label=255, return_test=False,
                 fixed_volume_space=False, max_volume_space=(50, 50, 1.5), min_volume_space=(-50, -50, -3)):
        'Initialization'
        self.point_cloud_dataset = in_dataset
        self.grid_size = np.asarray(grid_size)
        self.rotate_aug = rotate_aug
        self.ignore_label = ignore_label
        self.return_test = return_test
        self.flip_aug = flip_aug
        self.fixed_volume_space = fixed_volume_space
        self.max_volume_space = max_volume_space
        self.min_volume_space = min_volume_space

    def __len__(self):
        'Denotes the total number of samples'
        return len(self.point_cloud_dataset)

    def __getitem__(self, index):
        'Generates one sample of data'
        data = self.point_cloud_dataset[index]
        if len(data) == 2:
            xyz, labels = data
        elif len(data) == 3:
            xyz, labels, sig = data
            if len(sig.shape) == 2: sig = np.squeeze(sig)
        else:
            raise Exception('Return invalid data tuple')

        # random data augmentation by rotation
        if self.rotate_aug:
            rotate_rad = np.deg2rad(np.random.random() * 360)
            c, s = np.cos(rotate_rad), np.sin(rotate_rad)
            j = np.matrix([[c, s], [-s, c]])
            xyz[:, :2] = np.dot(xyz[:, :2], j)

        # random data augmentation by flip x , y or x+y
        if self.flip_aug:
            flip_type = np.random.choice(4, 1)
            if flip_type == 1:
                xyz[:, 0] = -xyz[:, 0]
            elif flip_type == 2:
                xyz[:, 1] = -xyz[:, 1]
            elif flip_type == 3:
                xyz[:, :2] = -xyz[:, :2]

        max_bound = np.percentile(xyz, 100, axis=0)
        min_bound = np.percentile(xyz, 0, axis=0)

        if self.fixed_volume_space:
            max_bound = np.asarray(self.max_volume_space)
            min_bound = np.asarray(self.min_volume_space)

        # get grid index
        crop_range = max_bound - min_bound
        cur_grid_size = self.grid_size

        intervals = crop_range / (cur_grid_size - 1)
        if (intervals == 0).any(): print("Zero interval!")

        grid_ind = (np.floor((np.clip(xyz, min_bound, max_bound) - min_bound) / intervals)).astype(np.int)

        # process voxel position
        voxel_position = np.zeros(self.grid_size, dtype=np.float32)
        dim_array = np.ones(len(self.grid_size) + 1, int)
        dim_array[0] = -1
        voxel_position = np.indices(self.grid_size) * intervals.reshape(dim_array) + min_bound.reshape(dim_array)

        # process labels
        processed_label = np.ones(self.grid_size, dtype=np.uint8) * self.ignore_label
        label_voxel_pair = np.concatenate([grid_ind, labels], axis=1)
        label_voxel_pair = label_voxel_pair[np.lexsort((grid_ind[:, 0], grid_ind[:, 1], grid_ind[:, 2])), :]
        processed_label = nb_process_label(np.copy(processed_label), label_voxel_pair)

        data_tuple = (voxel_position, processed_label)

        # center data on each voxel for PTnet
        voxel_centers = (grid_ind.astype(np.float32) + 0.5) * intervals + min_bound
        return_xyz = xyz - voxel_centers
        return_xyz = np.concatenate((return_xyz, xyz), axis=1)

        if len(data) == 2:
            return_fea = return_xyz
        elif len(data) == 3:
            return_fea = np.concatenate((return_xyz, sig[..., np.newaxis]), axis=1)

        if self.return_test:
            data_tuple += (grid_ind, labels, return_fea, index)
        else:
            data_tuple += (grid_ind, labels, return_fea)
        return data_tuple

# transformation between Cartesian coordinates and polar coordinates
def cart2polar(input_xyz):
    rho = np.sqrt(input_xyz[:, 0] ** 2 + input_xyz[:, 1] ** 2)
    phi = np.arctan2(input_xyz[:, 1], input_xyz[:, 0])
    return np.stack((rho, phi, input_xyz[:, 2]), axis=1)


def polar2cat(input_xyz_polar):
    # print(input_xyz_polar.shape)
    x = input_xyz_polar[0] * np.cos(input_xyz_polar[1])
    y = input_xyz_polar[0] * np.sin(input_xyz_polar[1])
    return np.stack((x, y, input_xyz_polar[2]), axis=0)


class SemanticKittiCylinder(data.Dataset):
    def __init__(self, in_dataset, grid_size = (480, 360, 32), rotate_aug=False, flip_aug=False, ignore_label=255, return_test=False,
                 fixed_volume_space=False, max_volume_space=(50, np.pi, 2), min_volume_space=(0, -np.pi, -4),
                 scale_aug=False,
                 transform_aug=False, trans_std=(0.1, 0.1, 0.1),
                 min_rad=-np.pi / 4, max_rad=np.pi / 4):
        self.point_cloud_dataset = in_dataset
        self.grid_size = np.asarray(grid_size)
        self.rotate_aug = rotate_aug
        self.flip_aug = flip_aug
        self.scale_aug = scale_aug
        self.ignore_label = ignore_label
        self.return_test = return_test
        self.fixed_volume_space = fixed_volume_space
        self.max_volume_space = max_volume_space
        self.min_volume_space = min_volume_space
        self.transform = transform_aug
        self.trans_std = trans_std
        self.noise_rotation = np.random.uniform(min_rad, max_rad)

    def __len__(self):
        'Denotes the total number of samples'
        return len(self.point_cloud_dataset)

    def rotation_points_single_angle(self, points, angle, axis=0):
        # points: [N, 3]
        rot_sin = np.sin(angle)
        rot_cos = np.cos(angle)
        if axis == 1:
            rot_mat_T = np.array(
                [[rot_cos, 0, -rot_sin], [0, 1, 0], [rot_sin, 0, rot_cos]],
                dtype=points.dtype)
        elif axis == 2 or axis == -1:
            rot_mat_T = np.array(
                [[rot_cos, -rot_sin, 0], [rot_sin, rot_cos, 0], [0, 0, 1]],
                dtype=points.dtype)
        elif axis == 0:
            rot_mat_T = np.array(
                [[1, 0, 0], [0, rot_cos, -rot_sin], [0, rot_sin, rot_cos]],
                dtype=points.dtype)
        else:
            raise ValueError("axis should in range")

        return points @ rot_mat_T

    def __getitem__(self, index):
        'Generates one sample of data'
        data = self.point_cloud_dataset[index]
        if len(data) == 2:
            xyz, labels = data
        elif len(data) == 3:
            xyz, labels, sig = data
            if len(sig.shape) == 2: sig = np.squeeze(sig)
        else:
            raise Exception('Return invalid data tuple')

        # random data augmentation by rotation
        if self.rotate_aug:
            rotate_rad = np.deg2rad(np.random.random() * 90) - np.pi / 4
            c, s = np.cos(rotate_rad), np.sin(rotate_rad)
            j = np.matrix([[c, s], [-s, c]])
            xyz[:, :2] = np.dot(xyz[:, :2], j)

        # random data augmentation by flip x , y or x+y
        if self.flip_aug:
            flip_type = np.random.choice(4, 1)
            if flip_type == 1:
                xyz[:, 0] = -xyz[:, 0]
            elif flip_type == 2:
                xyz[:, 1] = -xyz[:, 1]
            elif flip_type == 3:
                xyz[:, :2] = -xyz[:, :2]
        if self.scale_aug:
            noise_scale = np.random.uniform(0.95, 1.05)
            xyz[:, 0] = noise_scale * xyz[:, 0]
            xyz[:, 1] = noise_scale * xyz[:, 1]
        # convert coordinate into polar coordinates

        if self.transform:
            noise_translate = np.array([np.random.normal(0, self.trans_std[0], 1),
                                        np.random.normal(0, self.trans_std[1], 1),
                                        np.random.normal(0, self.trans_std[2], 1)]).T

            xyz[:, 0:3] += noise_translate

        xyz_pol = cart2polar(xyz)

        max_bound_r = np.percentile(xyz_pol[:, 0], 100, axis=0)
        min_bound_r = np.percentile(xyz_pol[:, 0], 0, axis=0)
        max_bound = np.max(xyz_pol[:, 1:], axis=0)
        min_bound = np.min(xyz_pol[:, 1:], axis=0)
        max_bound = np.concatenate(([max_bound_r], max_bound))
        min_bound = np.concatenate(([min_bound_r], min_bound))
        if self.fixed_volume_space:
            max_bound = np.asarray(self.max_volume_space)
            min_bound = np.asarray(self.min_volume_space)
        # get grid index
        crop_range = max_bound - min_bound
        cur_grid_size = self.grid_size
        intervals = crop_range / (cur_grid_size - 1)

        if (intervals == 0).any(): print("Zero interval!")
        grid_ind = (np.floor((np.clip(xyz_pol, min_bound, max_bound) - min_bound) / intervals)).astype(np.int)

        voxel_position = np.zeros(self.grid_size, dtype=np.float32)
        dim_array = np.ones(len(self.grid_size) + 1, int)
        dim_array[0] = -1
        voxel_position = np.indices(self.grid_size) * intervals.reshape(dim_array) + min_bound.reshape(dim_array)
        voxel_position = polar2cat(voxel_position)

        processed_label = np.ones(self.grid_size, dtype=np.uint8) * self.ignore_label
        label_voxel_pair = np.concatenate([grid_ind, labels], axis=1)
        label_voxel_pair = label_voxel_pair[np.lexsort((grid_ind[:, 0], grid_ind[:, 1], grid_ind[:, 2])), :]
        processed_label = nb_process_label(np.copy(processed_label), label_voxel_pair)
        data_tuple = (voxel_position, processed_label)

        # center data on each voxel for PTnet
        voxel_centers = (grid_ind.astype(np.float32) + 0.5) * intervals + min_bound
        return_xyz = xyz_pol - voxel_centers
        return_xyz = np.concatenate((return_xyz, xyz_pol, xyz[:, :2]), axis=1)

        if len(data) == 2:
            return_fea = return_xyz
        elif len(data) == 3:
            return_fea = np.concatenate((return_xyz, sig[..., np.newaxis]), axis=1)

        if self.return_test:
            data_tuple += (grid_ind, labels, return_fea, index)
        else:
            data_tuple += (grid_ind, labels, return_fea)
        return data_tuple


class SemanticKittiPolar(data.Dataset):
    def __init__(self, in_dataset, grid_size = (480, 360, 32), rotate_aug=False, flip_aug=False, ignore_label=255, return_test=False,
                 fixed_volume_space=False, max_volume_space=(50, np.pi, 2), min_volume_space=(0, -np.pi, -4),
                 scale_aug=False):
        self.point_cloud_dataset = in_dataset
        self.grid_size = np.asarray(grid_size)
        self.rotate_aug = rotate_aug
        self.flip_aug = flip_aug
        self.scale_aug = scale_aug
        self.ignore_label = ignore_label
        self.return_test = return_test
        self.fixed_volume_space = fixed_volume_space
        self.max_volume_space = max_volume_space
        self.min_volume_space = min_volume_space

    def __len__(self):
        'Denotes the total number of samples'
        return len(self.point_cloud_dataset)

    def __getitem__(self, index):
        'Generates one sample of data'
        data = self.point_cloud_dataset[index]
        if len(data) == 2:
            xyz, labels = data
        elif len(data) == 3:
            xyz, labels, sig = data
            if len(sig.shape) == 2:
                sig = np.squeeze(sig)
        else:
            raise Exception('Return invalid data tuple')

        # random data augmentation by rotation
        if self.rotate_aug:
            rotate_rad = np.deg2rad(np.random.random() * 45) - np.pi / 8
            c, s = np.cos(rotate_rad), np.sin(rotate_rad)
            j = np.matrix([[c, s], [-s, c]])
            xyz[:, :2] = np.dot(xyz[:, :2], j)

        # random data augmentation by flip x , y or x+y
        if self.flip_aug:
            flip_type = np.random.choice(4, 1)
            if flip_type == 1:
                xyz[:, 0] = -xyz[:, 0]
            elif flip_type == 2:
                xyz[:, 1] = -xyz[:, 1]
            elif flip_type == 3:
                xyz[:, :2] = -xyz[:, :2]
        if self.scale_aug:
            noise_scale = np.random.uniform(0.95, 1.05)
            xyz[:, 0] = noise_scale * xyz[:, 0]
            xyz[:, 1] = noise_scale * xyz[:, 1]
        xyz_pol = cart2polar(xyz)

        max_bound_r = np.percentile(xyz_pol[:, 0], 100, axis=0)
        min_bound_r = np.percentile(xyz_pol[:, 0], 0, axis=0)
        max_bound = np.max(xyz_pol[:, 1:], axis=0)
        min_bound = np.min(xyz_pol[:, 1:], axis=0)
        max_bound = np.concatenate(([max_bound_r], max_bound))
        min_bound = np.concatenate(([min_bound_r], min_bound))
        if self.fixed_volume_space:
            max_bound = np.asarray(self.max_volume_space)
            min_bound = np.asarray(self.min_volume_space)
        # get grid index
        crop_range = max_bound - min_bound
        cur_grid_size = self.grid_size
        intervals = crop_range / (cur_grid_size - 1)

        if (intervals == 0).any(): print("Zero interval!")
        grid_ind = (np.floor((np.clip(xyz_pol, min_bound, max_bound) - min_bound) / intervals)).astype(np.int)

        voxel_position = np.zeros(self.grid_size, dtype=np.float32)
        dim_array = np.ones(len(self.grid_size) + 1, int)
        dim_array[0] = -1
        voxel_position = np.indices(self.grid_size) * intervals.reshape(dim_array) + min_bound.reshape(dim_array)
        voxel_position = polar2cat(voxel_position)

        processed_label = np.ones(self.grid_size, dtype=np.uint8) * self.ignore_label
        label_voxel_pair = np.concatenate([grid_ind, labels], axis=1)
        label_voxel_pair = label_voxel_pair[np.lexsort((grid_ind[:, 0], grid_ind[:, 1], grid_ind[:, 2])), :]
        processed_label = nb_process_label(np.copy(processed_label), label_voxel_pair)
        data_tuple = (voxel_position, processed_label)

        # center data on each voxel for PTnet
        voxel_centers = (grid_ind.astype(np.float32) + 0.5) * intervals + min_bound
        return_xyz = xyz_pol - voxel_centers
        return_xyz = np.concatenate((return_xyz, xyz_pol, xyz[:, :2]), axis=1)

        if len(data) == 2:
            return_fea = return_xyz
        elif len(data) == 3:
            return_fea = np.concatenate((return_xyz, sig[..., np.newaxis]), axis=1)

        if self.return_test:
            data_tuple += (grid_ind, labels, return_fea, index)
        else:
            data_tuple += (grid_ind, labels, return_fea)

        return data_tuple


# @nb.jit('u1[:,:,:](u1[:,:,:],i8[:,:])', nopython=True, cache=True, parallel=False)
def nb_process_label(processed_label, sorted_label_voxel_pair):
    label_size = 256
    counter = np.zeros((label_size,), dtype=np.uint16)
    counter[sorted_label_voxel_pair[0, 3]] = 1
    cur_sear_ind = sorted_label_voxel_pair[0, :3]
    for i in range(1, sorted_label_voxel_pair.shape[0]):
        cur_ind = sorted_label_voxel_pair[i, :3]
        if not np.all(np.equal(cur_ind, cur_sear_ind)):
            processed_label[cur_sear_ind[0], cur_sear_ind[1], cur_sear_ind[2]] = np.argmax(counter)
            counter = np.zeros((label_size,), dtype=np.uint16)
            cur_sear_ind = cur_ind
        counter[sorted_label_voxel_pair[i, 3]] += 1
    processed_label[cur_sear_ind[0], cur_sear_ind[1], cur_sear_ind[2]] = np.argmax(counter)
    return processed_label


def collate_fn_BEV(data):
    data2stack = np.stack([d[0] for d in data]).astype(np.float32)
    label2stack = np.stack([d[1] for d in data]).astype(np.int)
    grid_ind_stack = [d[2] for d in data]
    point_label = [d[3] for d in data]
    xyz = [d[4] for d in data]
    return torch.from_numpy(data2stack), torch.from_numpy(label2stack), grid_ind_stack, point_label, xyz


def collate_fn_BEV_test(data):
    data2stack = np.stack([d[0] for d in data]).astype(np.float32)
    label2stack = np.stack([d[1] for d in data]).astype(np.int)
    grid_ind_stack = [d[2] for d in data]
    point_label = [d[3] for d in data]
    xyz = [d[4] for d in data]
    index = [d[5] for d in data]
    return torch.from_numpy(data2stack), torch.from_numpy(label2stack), grid_ind_stack, point_label, xyz, index




#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2020/6/4 17:25:47

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import inspect, sys, time, traceback, math, re, codecs
#===============================================================================
# 交互式运行 _lktest
#===============================================================================
FUNC_NAME = "_lktest"

__all__ = ['lktest_run']

def _time_run(kvfunc, catch_except=True):
    try:
        name = kvfunc[0]
        func = kvfunc[1]
        print( '%s==begin' % name)
        t0 = time.time()
        func()
    except SystemExit:
        exit(0)
    except Exception as e:
        if catch_except:
            print ( '\nexception in call function---->\n%s' % traceback.format_exc() )
        else:
            raise
    finally:
        elapse = time.time() - t0
        print ( '%s==end(time:%fs)\n' % (name, elapse) )

#从文件提取 _lktest名字,主要是用于排序
RE_PAT = re.compile(r"^def\s*(.*{})\(\):.*".format(FUNC_NAME))
def _find_file_lktests(filename):
    with codecs.open(filename, encoding='utf8') as f:
        for i,line in enumerate(f):
            res = RE_PAT.search(line)
            if res:
                yield i,res.group(1)

def lktest_run(run_opt='run_last', catch_except=True):
    '''
    :param run_opt: run_last/run_first/run_select/run_select_once/run_all/[name_lktest]
    :param catch_except:
    :return:
    '''

    # 获取所有 _lktest
    mainMod = sys.modules['__main__']
    kvList = [ (k,v) for k,v in inspect.getmembers(mainMod, inspect.isfunction)
               if k.endswith(FUNC_NAME) ]

    # 检测是否有测试用例
    if not kvList:
        print('there is no [name]{} here'.format(FUNC_NAME))
        return

    # 按照代码文本顺序排序
    kvOrder = {}
    for order, name in  _find_file_lktests(mainMod.__file__):
        kvOrder[name] = order
    kvList.sort(key = lambda pair: kvOrder[pair[0]])

    if run_opt == 'run_first':
        return _time_run(kvList[0], catch_except)
    elif run_opt == 'run_last':
        return _time_run(kvList[-1], catch_except)
    elif run_opt == 'run_all':
        for kv in kvList:
            _time_run(kv, catch_except)
    elif run_opt in ['run_select','run_select_once', '']:
        listLen = len(kvList)
        while True:
            w = math.ceil(math.log10(listLen))
            fmt = "<%0{}d>==%s".format(w)
            for i, kv in enumerate(kvList):
                print(fmt % (i, kv[0]))

            i = input('which test do you want to run , -1=quit , -2=run-all:')
            try:
                i = int(i)
            except:
                i = listLen - 1
            if i == -2:
                for kv in kvList:
                    _time_run(kv, catch_except)
                return
            elif i == -1:
                return
            else:
                i = (i + listLen) % listLen
                _time_run( kvList[i], catch_except)

            if run_opt == 'run_select_once':
                return
    else :
        if not isinstance(run_opt, list):
            run_list = [ run_opt ]
        else:
            run_list = run_opt

        for _run_op in run_list:
            for kv in kvList:
                if (isinstance(_run_op, str) and kv[0] == _run_op ) \
                        or (inspect.isfunction(_run_op) and kv[1] == _run_op):
                    _time_run(kv, catch_except)





#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/5/20 09:56:07

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
from collections import OrderedDict
import inspect
#===============================================================================
'''     
'''
#===============================================================================
__all__ = ['ModelClassSet', 'ModelDetectSet', 'ModelSegmentSet', 'model_class_set']


class ModelSet:
    '''基本的仓库
    '''
    def __init__(self):
        self.model_cls_set_ = OrderedDict()

    def createModel(self, name, **kws):
        return self.model_cls_set_[name](**kws)

    def __iter__(self):
        return iter(self.model_cls_set_.keys())

    def __len__(self):
        return len(self.model_cls_set_)

    def __contains__(self, item):
        return item in self.model_cls_set_

    def __getitem__(self, item):
        return self.model_cls_set_[item]

    def addModelCreator(self, creator, name = None):
        if name is None:
            name = creator.__name__
        if name in self.model_cls_set_:
            raise RuntimeError(f"Already exists: {name}")
        self.model_cls_set_[name] = creator

    def show(self):
        for k,v in self.model_cls_set_.items():
            print(f'{k}={v}')

class ModelClassSet(ModelSet):
    ''' 识别模型的仓库
    '''
    def __init__(self):
        super(ModelClassSet, self).__init__()

        import torchvision.models as m
        model_cls_list = [
            m.alexnet,
            m.convnext_tiny, m.convnext_small, m.convnext_base, m.convnext_large,
            m.densenet121, m.densenet169, m.densenet201, m.densenet161,
            m.efficientnet_b0, m.efficientnet_b1, m.efficientnet_b2, m.efficientnet_b3, m.efficientnet_b4, m.efficientnet_b5, m.efficientnet_b6, m.efficientnet_b7,
            m.googlenet,
            m.inception_v3,
            m.mnasnet0_5, m.mnasnet0_75, m.mnasnet1_0, m.mnasnet1_3,
            m.mobilenet_v2, m.mobilenet_v3_large, m.mobilenet_v3_small,
            m.regnet_y_400mf,m.regnet_y_800mf,m.regnet_y_1_6gf,m.regnet_y_3_2gf,m.regnet_y_8gf,m.regnet_y_16gf,m.regnet_y_32gf,m.regnet_y_128gf,m.regnet_x_400mf,m.regnet_x_800mf,m.regnet_x_1_6gf,m.regnet_x_3_2gf,m.regnet_x_8gf,m.regnet_x_16gf,m.regnet_x_32gf,
            m.resnet18, m.resnet34, m.resnet50, m.resnet101, m.resnet152, m.resnext50_32x4d, m.resnext101_32x8d, m.wide_resnet50_2, m.wide_resnet101_2,
            m.shufflenet_v2_x0_5, m.shufflenet_v2_x1_0, m.shufflenet_v2_x1_5, m.shufflenet_v2_x2_0,
            m.squeezenet1_0, m.squeezenet1_1,
            m.vgg11, m.vgg11_bn, m.vgg13, m.vgg13_bn, m.vgg16, m.vgg16_bn, m.vgg19_bn, m.vgg19,
            m.vit_b_16, m.vit_b_32, m.vit_l_16, m.vit_l_32,
        ]
        for model_cls in model_cls_list:
            self.addModelCreator(model_cls)

class ModelDetectSet(ModelSet):
    '''检测模型仓库
    '''
    def __init__(self):
        super(ModelDetectSet, self).__init__()

        import torchvision.models.detection as m
        model_cls_list = [
            m.fasterrcnn_resnet50_fpn, m.fasterrcnn_mobilenet_v3_large_320_fpn, m.fasterrcnn_mobilenet_v3_large_fpn,
            m.fcos_resnet50_fpn,
            m.keypointrcnn_resnet50_fpn,
            m.maskrcnn_resnet50_fpn,
            m.retinanet_resnet50_fpn,
            m.ssd300_vgg16,
            m.ssdlite320_mobilenet_v3_large,
        ]
        for model_cls in model_cls_list:
            self.addModelCreator(model_cls)

    def createModel(self, name, **kws):
        func = self.model_cls_set_[name]
        if 'pretrained_backbone' in inspect.getcallargs(func):
            kws['pretrained_backbone'] = kws.pop('pretrained_backbone', False)
        return func(**kws)

class ModelSegmentSet(ModelSet):
    '''检测模型仓库
    '''
    def __init__(self):
        super(ModelSegmentSet, self).__init__()

        import torchvision.models.segmentation as m
        model_cls_list = [
            m.deeplabv3_resnet50, m.deeplabv3_resnet101, m.deeplabv3_mobilenet_v3_large,
            m.fcn_resnet50, m.fcn_resnet101,
            m.lraspp_mobilenet_v3_large,
        ]
        for model_cls in model_cls_list:
            self.addModelCreator(model_cls)

    def createModel(self, name, **kws):
        func = self.model_cls_set_[name]
        if 'pretrained_backbone' in inspect.getcallargs(func):
            kws['pretrained_backbone'] = kws.pop('pretrained_backbone', False)
        return func(**kws)


model_class_set = ModelClassSet()




#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/3/7 15:44:41

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
#===============================================================================
# 
#===============================================================================
__all__ = [
    'farthest_point_sample',
    'pc_normalize',
    'normalize_data',
    'shuffle_data',

]

def normalize_data(batch_data):
    """ Normalize the batch data, use coordinates of the block centered at origin,
        Input:
            BxNxC array
        Output:
            BxNxC array
    """
    B, N, C = batch_data.shape
    normal_data = np.zeros((B, N, C))
    for b in range(B):
        pc = batch_data[b]
        centroid = np.mean(pc, axis=0)
        pc = pc - centroid
        m = np.max(np.sqrt(np.sum(pc ** 2, axis=1)))
        pc = pc / m
        normal_data[b] = pc
    return normal_data


def shuffle_data(data, labels):
    """ Shuffle data and labels.
        Input:
          data: B,N,... numpy array
          label: B,... numpy array
        Return:
          shuffled data, label and shuffle indices
    """
    idx = np.arange(len(labels))
    np.random.shuffle(idx)
    return data[idx, ...], labels[idx], idx

def shuffle_points(batch_data):
    """ Shuffle orders of points in each point cloud -- changes FPS behavior.
        Use the same shuffling idx for the entire batch.
        Input:
            BxNxC array
        Output:
            BxNxC array
    """
    idx = np.arange(batch_data.shape[1])
    np.random.shuffle(idx)
    return batch_data[:,idx,:]

def rotate_point_cloud(batch_data):
    """ Randomly rotate the point clouds to augument the dataset
        rotation is per shape based along up direction
        Input:
          BxNx3 array, original batch of point clouds
        Return:
          BxNx3 array, rotated batch of point clouds
    """
    rotated_data = np.zeros(batch_data.shape, dtype=np.float32)
    for k in range(batch_data.shape[0]):
        rotation_angle = np.random.uniform() * 2 * np.pi
        cosval = np.cos(rotation_angle)
        sinval = np.sin(rotation_angle)
        rotation_matrix = np.array([[cosval, 0, sinval],
                                    [0, 1, 0],
                                    [-sinval, 0, cosval]])
        shape_pc = batch_data[k, ...]
        rotated_data[k, ...] = np.dot(shape_pc.reshape((-1, 3)), rotation_matrix)
    return rotated_data

def rotate_point_cloud_z(batch_data):
    """ Randomly rotate the point clouds to augument the dataset
        rotation is per shape based along up direction
        Input:
          BxNx3 array, original batch of point clouds
        Return:
          BxNx3 array, rotated batch of point clouds
    """
    rotated_data = np.zeros(batch_data.shape, dtype=np.float32)
    for k in range(batch_data.shape[0]):
        rotation_angle = np.random.uniform() * 2 * np.pi
        cosval = np.cos(rotation_angle)
        sinval = np.sin(rotation_angle)
        rotation_matrix = np.array([[cosval, sinval, 0],
                                    [-sinval, cosval, 0],
                                    [0, 0, 1]])
        shape_pc = batch_data[k, ...]
        rotated_data[k, ...] = np.dot(shape_pc.reshape((-1, 3)), rotation_matrix)
    return rotated_data

def rotate_point_cloud_with_normal(batch_xyz_normal):
    ''' Randomly rotate XYZ, normal point cloud.
        Input:
            batch_xyz_normal: B,N,6, first three channels are XYZ, last 3 all normal
        Output:
            B,N,6, rotated XYZ, normal point cloud
    '''
    for k in range(batch_xyz_normal.shape[0]):
        rotation_angle = np.random.uniform() * 2 * np.pi
        cosval = np.cos(rotation_angle)
        sinval = np.sin(rotation_angle)
        rotation_matrix = np.array([[cosval, 0, sinval],
                                    [0, 1, 0],
                                    [-sinval, 0, cosval]])
        shape_pc = batch_xyz_normal[k,:,0:3]
        shape_normal = batch_xyz_normal[k,:,3:6]
        batch_xyz_normal[k,:,0:3] = np.dot(shape_pc.reshape((-1, 3)), rotation_matrix)
        batch_xyz_normal[k,:,3:6] = np.dot(shape_normal.reshape((-1, 3)), rotation_matrix)
    return batch_xyz_normal

def rotate_perturbation_point_cloud_with_normal(batch_data, angle_sigma=0.06, angle_clip=0.18):
    """ Randomly perturb the point clouds by small rotations
        Input:
          BxNx6 array, original batch of point clouds and point normals
        Return:
          BxNx3 array, rotated batch of point clouds
    """
    rotated_data = np.zeros(batch_data.shape, dtype=np.float32)
    for k in range(batch_data.shape[0]):
        angles = np.clip(angle_sigma*np.random.randn(3), -angle_clip, angle_clip)
        Rx = np.array([[1,0,0],
                       [0,np.cos(angles[0]),-np.sin(angles[0])],
                       [0,np.sin(angles[0]),np.cos(angles[0])]])
        Ry = np.array([[np.cos(angles[1]),0,np.sin(angles[1])],
                       [0,1,0],
                       [-np.sin(angles[1]),0,np.cos(angles[1])]])
        Rz = np.array([[np.cos(angles[2]),-np.sin(angles[2]),0],
                       [np.sin(angles[2]),np.cos(angles[2]),0],
                       [0,0,1]])
        R = np.dot(Rz, np.dot(Ry,Rx))
        shape_pc = batch_data[k,:,0:3]
        shape_normal = batch_data[k,:,3:6]
        rotated_data[k,:,0:3] = np.dot(shape_pc.reshape((-1, 3)), R)
        rotated_data[k,:,3:6] = np.dot(shape_normal.reshape((-1, 3)), R)
    return rotated_data


def rotate_point_cloud_by_angle(batch_data, rotation_angle):
    """ Rotate the point cloud along up direction with certain angle.
        Input:
          BxNx3 array, original batch of point clouds
        Return:
          BxNx3 array, rotated batch of point clouds
    """
    rotated_data = np.zeros(batch_data.shape, dtype=np.float32)
    for k in range(batch_data.shape[0]):
        #rotation_angle = np.random.uniform() * 2 * np.pi
        cosval = np.cos(rotation_angle)
        sinval = np.sin(rotation_angle)
        rotation_matrix = np.array([[cosval, 0, sinval],
                                    [0, 1, 0],
                                    [-sinval, 0, cosval]])
        shape_pc = batch_data[k,:,0:3]
        rotated_data[k,:,0:3] = np.dot(shape_pc.reshape((-1, 3)), rotation_matrix)
    return rotated_data

def rotate_point_cloud_by_angle_with_normal(batch_data, rotation_angle):
    """ Rotate the point cloud along up direction with certain angle.
        Input:
          BxNx6 array, original batch of point clouds with normal
          scalar, angle of rotation
        Return:
          BxNx6 array, rotated batch of point clouds iwth normal
    """
    rotated_data = np.zeros(batch_data.shape, dtype=np.float32)
    for k in range(batch_data.shape[0]):
        #rotation_angle = np.random.uniform() * 2 * np.pi
        cosval = np.cos(rotation_angle)
        sinval = np.sin(rotation_angle)
        rotation_matrix = np.array([[cosval, 0, sinval],
                                    [0, 1, 0],
                                    [-sinval, 0, cosval]])
        shape_pc = batch_data[k,:,0:3]
        shape_normal = batch_data[k,:,3:6]
        rotated_data[k,:,0:3] = np.dot(shape_pc.reshape((-1, 3)), rotation_matrix)
        rotated_data[k,:,3:6] = np.dot(shape_normal.reshape((-1,3)), rotation_matrix)
    return rotated_data



def rotate_perturbation_point_cloud(batch_data, angle_sigma=0.06, angle_clip=0.18):
    """ Randomly perturb the point clouds by small rotations
        Input:
          BxNx3 array, original batch of point clouds
        Return:
          BxNx3 array, rotated batch of point clouds
    """
    rotated_data = np.zeros(batch_data.shape, dtype=np.float32)
    for k in range(batch_data.shape[0]):
        angles = np.clip(angle_sigma*np.random.randn(3), -angle_clip, angle_clip)
        Rx = np.array([[1,0,0],
                       [0,np.cos(angles[0]),-np.sin(angles[0])],
                       [0,np.sin(angles[0]),np.cos(angles[0])]])
        Ry = np.array([[np.cos(angles[1]),0,np.sin(angles[1])],
                       [0,1,0],
                       [-np.sin(angles[1]),0,np.cos(angles[1])]])
        Rz = np.array([[np.cos(angles[2]),-np.sin(angles[2]),0],
                       [np.sin(angles[2]),np.cos(angles[2]),0],
                       [0,0,1]])
        R = np.dot(Rz, np.dot(Ry,Rx))
        shape_pc = batch_data[k, ...]
        rotated_data[k, ...] = np.dot(shape_pc.reshape((-1, 3)), R)
    return rotated_data


def jitter_point_cloud(batch_data, sigma=0.01, clip=0.05):
    """ Randomly jitter points. jittering is per point.
        Input:
          BxNx3 array, original batch of point clouds
        Return:
          BxNx3 array, jittered batch of point clouds
    """
    B, N, C = batch_data.shape
    assert(clip > 0)
    jittered_data = np.clip(sigma * np.random.randn(B, N, C), -1*clip, clip)
    jittered_data += batch_data
    return jittered_data

def shift_point_cloud(batch_data, shift_range=0.1):
    """ Randomly shift point cloud. Shift is per point cloud.
        Input:
          BxNx3 array, original batch of point clouds
        Return:
          BxNx3 array, shifted batch of point clouds
    """
    B, N, C = batch_data.shape
    shifts = np.random.uniform(-shift_range, shift_range, (B,3))
    for batch_index in range(B):
        batch_data[batch_index,:,:] += shifts[batch_index,:]
    return batch_data


def random_scale_point_cloud(batch_data, scale_low=0.8, scale_high=1.25):
    """ Randomly scale the point cloud. Scale is per point cloud.
        Input:
            BxNx3 array, original batch of point clouds
        Return:
            BxNx3 array, scaled batch of point clouds
    """
    B, N, C = batch_data.shape
    scales = np.random.uniform(scale_low, scale_high, B)
    for batch_index in range(B):
        batch_data[batch_index,:,:] *= scales[batch_index]
    return batch_data

def random_point_dropout(batch_pc, max_dropout_ratio=0.875):
    ''' batch_pc: BxNx3 '''
    for b in range(batch_pc.shape[0]):
        dropout_ratio =  np.random.random()*max_dropout_ratio # 0~0.875
        drop_idx = np.where(np.random.random((batch_pc.shape[1]))<=dropout_ratio)[0]
        if len(drop_idx)>0:
            batch_pc[b,drop_idx,:] = batch_pc[b,0,:] # set to the first point
    return batch_pc


# reference https://github.com/yanx27/Pointnet_Pointnet2_pytorch
def pc_normalize(pc):
    centroid = np.mean(pc, axis=0)
    pc = pc - centroid
    m = np.max(np.sqrt(np.sum(pc**2, axis=1)))
    pc = pc / m
    return pc

def square_distance(src, dst):
    """
    Calculate Euclid distance between each two points.
    src^T * dst = xn * xm + yn * ym + zn * zm；
    sum(src^2, dim=-1) = xn*xn + yn*yn + zn*zn;
    sum(dst^2, dim=-1) = xm*xm + ym*ym + zm*zm;
    dist = (xn - xm)^2 + (yn - ym)^2 + (zn - zm)^2
         = sum(src**2,dim=-1)+sum(dst**2,dim=-1)-2*src^T*dst
    Input:
        src: source points, [B, N, C]
        dst: target points, [B, M, C]
    Output:
        dist: per-point square distance, [B, N, M]
    """
    return torch.sum((src[:, :, None] - dst[:, None]) ** 2, dim=-1)


def index_points(points, idx):
    """
    Input:
        points: input points data, [B, N, C]
        idx: sample index data, [B, S, [K]]
    Return:
        new_points:, indexed points data, [B, S, [K], C]
    """
    raw_size = idx.size()
    idx = idx.reshape(raw_size[0], -1)
    res = torch.gather(points, 1, idx[..., None].expand(-1, -1, points.size(-1)))
    return res.reshape(*raw_size, -1)


def farthest_point_sample(xyz, npoint):
    """
    Input:
        xyz: pointcloud data, [B, N, 3]
        npoint: number of samples
    Return:
        centroids: sampled pointcloud index, [B, npoint]
    """
    device = xyz.device
    B, N, C = xyz.shape
    centroids = torch.zeros(B, npoint, dtype=torch.long).to(device)
    distance = torch.ones(B, N).to(device) * 1e10
    farthest = torch.randint(0, N, (B,), dtype=torch.long).to(device)
    batch_indices = torch.arange(B, dtype=torch.long).to(device)
    for i in range(npoint):
        centroids[:, i] = farthest
        centroid = xyz[batch_indices, farthest, :].view(B, 1, 3)
        dist = torch.sum((xyz - centroid) ** 2, -1)
        distance = torch.min(distance, dist)
        farthest = torch.max(distance, -1)[1]
    return centroids


def query_ball_point(radius, nsample, xyz, new_xyz):
    """
    Input:
        radius: local region radius
        nsample: max sample number in local region
        xyz: all points, [B, N, 3]
        new_xyz: query points, [B, S, 3]
    Return:
        group_idx: grouped points index, [B, S, nsample]
    """
    device = xyz.device
    B, N, C = xyz.shape
    _, S, _ = new_xyz.shape
    group_idx = torch.arange(N, dtype=torch.long).to(device).view(1, 1, N).repeat([B, S, 1])
    sqrdists = square_distance(new_xyz, xyz)
    group_idx[sqrdists > radius ** 2] = N
    group_idx = group_idx.sort(dim=-1)[0][:, :, :nsample]
    group_first = group_idx[:, :, 0].view(B, S, 1).repeat([1, 1, nsample])
    mask = group_idx == N
    group_idx[mask] = group_first[mask]
    return group_idx


def sample_and_group(npoint, radius, nsample, xyz, points, returnfps=False, knn=False):
    """
    Input:
        npoint:
        radius:
        nsample:
        xyz: input points position data, [B, N, 3]
        points: input points data, [B, N, D]
    Return:
        new_xyz: sampled points position data, [B, npoint, nsample, 3]
        new_points: sampled points data, [B, npoint, nsample, 3+D]
    """
    B, N, C = xyz.shape
    S = npoint
    fps_idx = farthest_point_sample(xyz, npoint) # [B, npoint]
    torch.cuda.empty_cache()
    new_xyz = index_points(xyz, fps_idx)
    torch.cuda.empty_cache()
    if knn:
        dists = square_distance(new_xyz, xyz)  # B x npoint x N
        idx = dists.argsort()[:, :, :nsample]  # B x npoint x K
    else:
        idx = query_ball_point(radius, nsample, xyz, new_xyz)
    torch.cuda.empty_cache()
    grouped_xyz = index_points(xyz, idx) # [B, npoint, nsample, C]
    torch.cuda.empty_cache()
    grouped_xyz_norm = grouped_xyz - new_xyz.view(B, S, 1, C)
    torch.cuda.empty_cache()

    if points is not None:
        grouped_points = index_points(points, idx)
        new_points = torch.cat([grouped_xyz_norm, grouped_points], dim=-1) # [B, npoint, nsample, C+D]
    else:
        new_points = grouped_xyz_norm
    if returnfps:
        return new_xyz, new_points, grouped_xyz, fps_idx
    else:
        return new_xyz, new_points


def sample_and_group_all(xyz, points):
    """
    Input:
        xyz: input points position data, [B, N, 3]
        points: input points data, [B, N, D]
    Return:
        new_xyz: sampled points position data, [B, 1, 3]
        new_points: sampled points data, [B, 1, N, 3+D]
    """
    device = xyz.device
    B, N, C = xyz.shape
    new_xyz = torch.zeros(B, 1, C).to(device)
    grouped_xyz = xyz.view(B, 1, N, C)
    if points is not None:
        new_points = torch.cat([grouped_xyz, points.view(B, 1, N, -1)], dim=-1)
    else:
        new_points = grouped_xyz
    return new_xyz, new_points


class PointNetSetAbstraction(nn.Module):
    def __init__(self, npoint, radius, nsample, in_channel, mlp, group_all, knn=False):
        super(PointNetSetAbstraction, self).__init__()
        self.npoint = npoint
        self.radius = radius
        self.nsample = nsample
        self.knn = knn
        self.mlp_convs = nn.ModuleList()
        self.mlp_bns = nn.ModuleList()
        last_channel = in_channel
        for out_channel in mlp:
            self.mlp_convs.append(nn.Conv2d(last_channel, out_channel, 1))
            self.mlp_bns.append(nn.BatchNorm2d(out_channel))
            last_channel = out_channel
        self.group_all = group_all

    def forward(self, xyz, points):
        """
        Input:
            xyz: input points position data, [B, N, C]
            points: input points data, [B, N, C]
        Return:
            new_xyz: sampled points position data, [B, S, C]
            new_points_concat: sample points feature data, [B, S, D']
        """
        if self.group_all:
            new_xyz, new_points = sample_and_group_all(xyz, points)
        else:
            new_xyz, new_points = sample_and_group(self.npoint, self.radius, self.nsample, xyz, points, knn=self.knn)
        # new_xyz: sampled points position data, [B, npoint, C]
        # new_points: sampled points data, [B, npoint, nsample, C+D]
        new_points = new_points.permute(0, 3, 2, 1) # [B, C+D, nsample,npoint]
        for i, conv in enumerate(self.mlp_convs):
            bn = self.mlp_bns[i]
            new_points =  F.relu(bn(conv(new_points)))

        new_points = torch.max(new_points, 2)[0].transpose(1, 2)
        return new_xyz, new_points


class PointNetSetAbstractionMsg(nn.Module):
    def __init__(self, npoint, radius_list, nsample_list, in_channel, mlp_list, knn=False):
        super(PointNetSetAbstractionMsg, self).__init__()
        self.npoint = npoint
        self.radius_list = radius_list
        self.nsample_list = nsample_list
        self.knn = knn
        self.conv_blocks = nn.ModuleList()
        self.bn_blocks = nn.ModuleList()
        for i in range(len(mlp_list)):
            convs = nn.ModuleList()
            bns = nn.ModuleList()
            last_channel = in_channel + 3
            for out_channel in mlp_list[i]:
                convs.append(nn.Conv2d(last_channel, out_channel, 1))
                bns.append(nn.BatchNorm2d(out_channel))
                last_channel = out_channel
            self.conv_blocks.append(convs)
            self.bn_blocks.append(bns)

    def forward(self, xyz, points, seed_idx=None):
        """
        Input:
            xyz: input points position data, [B, C, N]
            points: input points data, [B, D, N]
        Return:
            new_xyz: sampled points position data, [B, C, S]
            new_points_concat: sample points feature data, [B, D', S]
        """

        B, N, C = xyz.shape
        S = self.npoint
        new_xyz = index_points(xyz, farthest_point_sample(xyz, S) if seed_idx is None else seed_idx)
        new_points_list = []
        for i, radius in enumerate(self.radius_list):
            K = self.nsample_list[i]
            if self.knn:
                dists = square_distance(new_xyz, xyz)  # B x npoint x N
                group_idx = dists.argsort()[:, :, :K]  # B x npoint x K
            else:
                group_idx = query_ball_point(radius, K, xyz, new_xyz)
            grouped_xyz = index_points(xyz, group_idx)
            grouped_xyz -= new_xyz.view(B, S, 1, C)
            if points is not None:
                grouped_points = index_points(points, group_idx)
                grouped_points = torch.cat([grouped_points, grouped_xyz], dim=-1)
            else:
                grouped_points = grouped_xyz

            grouped_points = grouped_points.permute(0, 3, 2, 1)  # [B, D, K, S]
            for j in range(len(self.conv_blocks[i])):
                conv = self.conv_blocks[i][j]
                bn = self.bn_blocks[i][j]
                grouped_points =  F.relu(bn(conv(grouped_points)))
            new_points = torch.max(grouped_points, 2)[0]  # [B, D', S]
            new_points_list.append(new_points)

        new_points_concat = torch.cat(new_points_list, dim=1).transpose(1, 2)
        return new_xyz, new_points_concat


# NoteL this function swaps N and C
class PointNetFeaturePropagation(nn.Module):
    def __init__(self, in_channel, mlp):
        super(PointNetFeaturePropagation, self).__init__()
        self.mlp_convs = nn.ModuleList()
        self.mlp_bns = nn.ModuleList()
        last_channel = in_channel
        for out_channel in mlp:
            self.mlp_convs.append(nn.Conv1d(last_channel, out_channel, 1))
            self.mlp_bns.append(nn.BatchNorm1d(out_channel))
            last_channel = out_channel

    def forward(self, xyz1, xyz2, points1, points2):
        """
        Input:
            xyz1: input points position data, [B, C, N]
            xyz2: sampled input points position data, [B, C, S]
            points1: input points data, [B, D, N]
            points2: input points data, [B, D, S]
        Return:
            new_points: upsampled points data, [B, D', N]
        """
        xyz1 = xyz1.permute(0, 2, 1)
        xyz2 = xyz2.permute(0, 2, 1)

        points2 = points2.permute(0, 2, 1)
        B, N, C = xyz1.shape
        _, S, _ = xyz2.shape

        if S == 1:
            interpolated_points = points2.repeat(1, N, 1)
        else:
            dists = square_distance(xyz1, xyz2)
            dists, idx = dists.sort(dim=-1)
            dists, idx = dists[:, :, :3], idx[:, :, :3]  # [B, N, 3]

            dist_recip = 1.0 / (dists + 1e-8)
            norm = torch.sum(dist_recip, dim=2, keepdim=True)
            weight = dist_recip / norm
            interpolated_points = torch.sum(index_points(points2, idx) * weight.view(B, N, 3, 1), dim=2)

        if points1 is not None:
            points1 = points1.permute(0, 2, 1)
            new_points = torch.cat([points1, interpolated_points], dim=-1)
        else:
            new_points = interpolated_points

        new_points = new_points.permute(0, 2, 1)
        for i, conv in enumerate(self.mlp_convs):
            bn = self.mlp_bns[i]
            new_points = F.relu(bn(conv(new_points)))
        return new_points


#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/5/26 11:04:35

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import types
from pathlib import Path
import cv2, numpy as np
from laok.cv2d.gui import show, show_wait
#===============================================================================
'''     
'''
#===============================================================================
__all__ = ['read_smaple_img', 'run_sample_image']

def read_smaple_img(is_color=True):
    fPath = Path(__file__).parent.joinpath('lena.jpg')
    flags = cv2.IMREAD_COLOR if is_color else cv2.IMREAD_GRAYSCALE
    return cv2.imdecode(np.fromfile(fPath, dtype=np.uint8), flags=flags)

def _get_name(trans):
    if hasattr(trans, '__name__'):
        return trans.__name__
    if hasattr(trans, 'func'):
        return trans.func.__name__
    if hasattr(trans, '__class__'):
        return trans.__class__.__name__

def run_sample_image(trans, is_color=True):
    img = read_smaple_img(is_color)
    img_trans = trans(img)
    show(img_trans, _get_name(trans))
    show_wait(img)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/4/29 08:10:41

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import numpy as np
from ..statis import cld_bounding_box
#===============================================================================
'''     
'''
#===============================================================================
__all__ = ['proj_img_x_div', 'proj_img_y_div', 'proj_img_z_div']

def proj_img_x_div(cld, div=200):
    ind = 0
    minPt, maxPt = cld_bounding_box(cld)
    resolution = (maxPt - minPt) / (div - 1)
    resolution[ind] = 1
    cldXYZ = (cld[:, 0:3] - minPt) / resolution
    _fill = minPt[ind] - (maxPt[ind] - minPt[ind])
    img = np.full((div, div), fill_value=_fill, dtype=np.float32)
    for xyz in cldXYZ:
        y = int(xyz[1])
        z = int(xyz[2])
        img[z,y] = xyz[0]
    return img

def proj_img_z_div(cld, div=200):
    ind = 2
    minPt, maxPt = cld_bounding_box(cld)
    resolution = (maxPt - minPt) / (div - 1)
    resolution[ind] = 1
    cldXYZ = (cld[:, 0:3] - minPt) / resolution
    _fill = minPt[ind] - (maxPt[ind] - minPt[ind])

    img = np.full((div, div), fill_value=_fill, dtype=np.float32)
    for xyz in cldXYZ:
        x = int(xyz[0])
        y = int(xyz[1])
        img[y, x] = xyz[2]
    return img

def proj_img_y_div(cld, div=200):
    ind = 1
    minPt, maxPt = cld_bounding_box(cld)
    resolution = (maxPt - minPt) / (div - 1)
    resolution[ind] = 1
    cldXYZ = (cld[:, 0:3] - minPt) / resolution
    _fill = minPt[ind] - (maxPt[ind] - minPt[ind])

    img = np.full((div, div), fill_value=_fill, dtype=np.float32)
    for xyz in cldXYZ:
        x = int(xyz[0])
        z = int(xyz[2])
        img[z, x] = xyz[1]
    return img


#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/5/25 21:57:25

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import cv2, numpy as np
#===============================================================================
'''     
'''
#===============================================================================
__all__ = ['edge_laplace', 'edge_canny',
           'edge_sobel_x', 'edge_sobel_y', 'edge_sobel']

def edge_laplace(img, ksize=1, scale=1, delta=0, borderType=cv2.BORDER_DEFAULT):
    return cv2.Laplacian(img, ddepth = -1, ksize=ksize, scale=scale, delta=delta, borderType=borderType)


def edge_canny(grayImg, threshold1 = 50,  threshold2 = 150, apertureSize=3, L2gradient=False):
    return cv2.Canny(grayImg, threshold1 = threshold1, threshold2 = threshold2, apertureSize=apertureSize, L2gradient=L2gradient)

def edge_sobel_x(img, ksize=0, scale=1, delta=0, borderType=cv2.BORDER_DEFAULT):
    return cv2.Sobel(img, cv2.CV_16S, dx=1, dy=0, ksize=ksize, scale=scale, delta=delta, borderType=borderType)

def edge_sobel_y(img, ksize=0, scale=1, delta=0, borderType=cv2.BORDER_DEFAULT):
    return cv2.Sobel(img, cv2.CV_16S, dx=0, dy=1, ksize=ksize, scale=scale, delta=delta, borderType=borderType)

def edge_sobel(img, res_type='none', ksize=0, scale=1, delta=0, borderType=cv2.BORDER_DEFAULT):
    ''' res_type: 'none', 'trunc', 'abs'
    '''
    x_img = cv2.Sobel(img, cv2.CV_16S, dx=1, dy=0, ksize=ksize, scale=scale, delta=delta, borderType=borderType)
    y_img = cv2.Sobel(img, cv2.CV_16S, dx=0, dy=1, ksize=ksize, scale=scale, delta=delta, borderType=borderType)
    if res_type == 'trunc':
        res = np.clip(x_img, 0, 255)  + np.clip(y_img, 0, 255)
    elif res_type == 'abs':
        res = np.abs(x_img) + np.abs(y_img)
    else:
        res = x_img + y_img
    return cv2.normalize(res, dst=res, alpha=0, beta=255, norm_type=cv2.NORM_MINMAX, dtype=cv2.CV_8U)


#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/4/20 20:58:51

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import torch
import torch.nn as nn
import torch.nn.functional as F
#===============================================================================
'''     
'''
#===============================================================================

__all__ = ['SiLU',
           'Hardswish',

           ]

# SiLU https://arxiv.org/pdf/1606.08415.pdf ----------------------------------------------------------------------------
class SiLU(nn.Module):  # export-friendly version of nn.SiLU()
    @staticmethod
    def forward(x):
        return x * torch.sigmoid(x)


class Hardswish(nn.Module):  # export-friendly version of nn.Hardswish()
    @staticmethod
    def forward(x):
        # return x * F.hardsigmoid(x)  # for torchscript and CoreML
        return x * F.hardtanh(x + 3, 0., 6.) / 6.  # for torchscript, CoreML and ONNX


class MemoryEfficientSwish(nn.Module):
    class F(torch.autograd.Function):
        @staticmethod
        def forward(ctx, x):
            ctx.save_for_backward(x)
            return x * torch.sigmoid(x)

        @staticmethod
        def backward(ctx, grad_output):
            x = ctx.saved_tensors[0]
            sx = torch.sigmoid(x)
            return grad_output * (sx * (1 + x * (1 - sx)))

    def forward(self, x):
        return self.F.apply(x)


# Mish https://github.com/digantamisra98/Mish --------------------------------------------------------------------------
class Mish(nn.Module):
    @staticmethod
    def forward(x):
        return x * F.softplus(x).tanh()


class MemoryEfficientMish(nn.Module):
    class F(torch.autograd.Function):
        @staticmethod
        def forward(ctx, x):
            ctx.save_for_backward(x)
            return x.mul(torch.tanh(F.softplus(x)))  # x * tanh(ln(1 + exp(x)))

        @staticmethod
        def backward(ctx, grad_output):
            x = ctx.saved_tensors[0]
            sx = torch.sigmoid(x)
            fx = F.softplus(x).tanh()
            return grad_output * (fx + x * sx * (1 - fx * fx))

    def forward(self, x):
        return self.F.apply(x)


# FReLU https://arxiv.org/abs/2007.11824 -------------------------------------------------------------------------------
class FReLU(nn.Module):
    def __init__(self, c1, k=3):  # ch_in, kernel
        super().__init__()
        self.conv = nn.Conv2d(c1, c1, k, 1, 1, groups=c1, bias=False)
        self.bn = nn.BatchNorm2d(c1)

    def forward(self, x):
        return torch.max(x, self.bn(self.conv(x)))



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/5/22 17:32:21

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import sys
import logging
import logging.handlers
from pathlib import Path
import os.path
#===============================================================================
'''     
'''
#===============================================================================
__all__ = ['get_logger']


def _add_handler_rotate_file(filename, fmt = None):
    handler = logging.handlers.RotatingFileHandler(
        filename,
        maxBytes=1024 * 1024 * 1024 * 1,
        backupCount=5,
    )
    if fmt:
        handler.setFormatter(fmt)
    return handler

def _add_handler_file(filename, fmt = None):
    handler = logging.FileHandler(filename)
    if fmt:
        handler.setFormatter(fmt)
    return handler

def _add_handler_console(fmt = None):
    handler = logging.StreamHandler()
    if fmt:
        handler.setFormatter(fmt)
    return handler

def get_logger(name=None, filename=None, fmt=None, level=None):
    if name is None:
        try:
            name = Path(sys.modules.get('__main__').__file__).stem
        except :
            name = "laok"
    log = logging.getLogger(name)

    if fmt is None:
        fmt = '[%(asctime)s][%(levelname)s]%(message)s'
    fmtter = logging.Formatter(fmt)

    if level is None:
        level = logging.DEBUG
    log.setLevel(level)

    log.addHandler(_add_handler_console(fmtter))

    if filename is not None:
        _, ext = os.path.splitext(filename)
        if not ext:
            filename = filename + ".log"
        log.addHandler(_add_handler_file(filename, fmtter))

    return log




#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/4/20 20:58:45

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import torch
import torch.nn as nn
import math
#===============================================================================
'''     
'''
#===============================================================================

class BCEBlurWithLogitsLoss(nn.Module):
    # BCEwithLogitLoss() with reduced missing label effects.
    def __init__(self, alpha=0.05):
        super(BCEBlurWithLogitsLoss, self).__init__()
        self.loss_fcn = nn.BCEWithLogitsLoss(reduction='none')  # must be nn.BCEWithLogitsLoss()
        self.alpha = alpha

    def forward(self, pred, true):
        loss = self.loss_fcn(pred, true)
        pred = torch.sigmoid(pred)  # prob from logits
        dx = pred - true  # reduce only missing label effects
        # dx = (pred - true).abs()  # reduce missing label and false label effects
        alpha_factor = 1 - torch.exp((dx - 1) / (self.alpha + 1e-4))
        loss *= alpha_factor
        return loss.mean()


class FocalLoss(nn.Module):
    # Wraps focal loss around existing loss_fcn(), i.e. criteria = FocalLoss(nn.BCEWithLogitsLoss(), gamma=1.5)
    def __init__(self, loss_fcn, gamma=1.5, alpha=0.25):
        super(FocalLoss, self).__init__()
        self.loss_fcn = loss_fcn  # must be nn.BCEWithLogitsLoss()
        self.gamma = gamma
        self.alpha = alpha
        self.reduction = loss_fcn.reduction
        self.loss_fcn.reduction = 'none'  # required to apply FL to each element

    def forward(self, pred, true):
        loss = self.loss_fcn(pred, true)
        # p_t = torch.exp(-loss)
        # loss *= self.alpha * (1.000001 - p_t) ** self.gamma  # non-zero power for gradient stability

        # TF implementation https://github.com/tensorflow/addons/blob/v0.7.1/tensorflow_addons/losses/focal_loss.py
        pred_prob = torch.sigmoid(pred)  # prob from logits
        p_t = true * pred_prob + (1 - true) * (1 - pred_prob)
        alpha_factor = true * self.alpha + (1 - true) * (1 - self.alpha)
        modulating_factor = (1.0 - p_t) ** self.gamma
        loss *= alpha_factor * modulating_factor

        if self.reduction == 'mean':
            return loss.mean()
        elif self.reduction == 'sum':
            return loss.sum()
        else:  # 'none'
            return loss


class QFocalLoss(nn.Module):
    # Wraps Quality focal loss around existing loss_fcn(), i.e. criteria = FocalLoss(nn.BCEWithLogitsLoss(), gamma=1.5)
    def __init__(self, loss_fcn, gamma=1.5, alpha=0.25):
        super(QFocalLoss, self).__init__()
        self.loss_fcn = loss_fcn  # must be nn.BCEWithLogitsLoss()
        self.gamma = gamma
        self.alpha = alpha
        self.reduction = loss_fcn.reduction
        self.loss_fcn.reduction = 'none'  # required to apply FL to each element

    def forward(self, pred, true):
        loss = self.loss_fcn(pred, true)

        pred_prob = torch.sigmoid(pred)  # prob from logits
        alpha_factor = true * self.alpha + (1 - true) * (1 - self.alpha)
        modulating_factor = torch.abs(true - pred_prob) ** self.gamma
        loss *= alpha_factor * modulating_factor

        if self.reduction == 'mean':
            return loss.mean()
        elif self.reduction == 'sum':
            return loss.sum()
        else:  # 'none'
            return loss

class WingLoss(nn.Module):
    def __init__(self, w=10, e=2):
        super(WingLoss, self).__init__()
        # https://arxiv.org/pdf/1711.06753v4.pdf   Figure 5
        self.w = w
        self.e = e
        self.C = self.w - self.w * math.log(1 + self.w / self.e)

    def forward(self, x, t, sigma=1):
        weight = torch.ones_like(t)
        weight[torch.where(t==-1)] = 0
        diff = weight * (x - t)
        abs_diff = diff.abs()
        flag = (abs_diff.data < self.w).float()
        y = flag * self.w * torch.log(1 + abs_diff / self.e) + (1 - flag) * (abs_diff - self.C)
        return y.sum()

class LandmarksLoss(nn.Module):
    # BCEwithLogitLoss() with reduced missing label effects.
    def __init__(self, alpha=1.0):
        super(LandmarksLoss, self).__init__()
        self.loss_fcn = WingLoss()#nn.SmoothL1Loss(reduction='sum')
        self.alpha = alpha

    def forward(self, pred, truel, mask):
        loss = self.loss_fcn(pred*mask, truel*mask)
        return loss / (torch.sum(mask) + 10e-14)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/5/26 00:43:37

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import cv2, numpy as np
#===============================================================================
'''     
'''
#===============================================================================
__all__ = ['filter_blur', 'filter_gauss', 'filter_bilateral', 'filter_median',
           ]

def filter_blur(img, ksize=(3,3), anchor=(-1,-1), borderType=cv2.BORDER_DEFAULT):
    return cv2.blur(img, ksize=ksize, anchor=anchor, borderType=borderType)

def filter_gauss(img, ksize, sigmaX, sigmaY=None, borderType=None):
    return cv2.GaussianBlur(img, ksize=ksize, sigmaX=sigmaX, sigmaY=sigmaY, borderType=borderType)

def filter_bilateral():
    pass

def filter_median():
    pass

def filter_meansift():
    pass

